#
# Copyright 2013 Radware LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib.api import attributes
from neutron_lib import context as ncontext
from neutron_lib import exceptions as lib_exc
from neutron_lib.exceptions import flavors as flav_exc
from neutron_lib.plugins import directory

from neutron.api.v2 import base as napi_base
from neutron.db import agentschedulers_db
from neutron.db import servicetype_db as st_db
from neutron import service
from neutron.services.flavors import flavors_plugin
from neutron.services import provider_configuration as pconf
from neutron.services import service_base
from neutron_lib import constants as n_constants
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants
from oslo_log import log as logging
from oslo_utils import encodeutils

from neutron_lbaas import agent_scheduler as agent_scheduler_v2
import neutron_lbaas.common.cert_manager
from neutron_lbaas.common import exceptions
from neutron_lbaas.common.tls_utils import cert_parser
from neutron_lbaas.db.loadbalancer import loadbalancer_dbv2 as ldbv2
from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.extensions import l7
from neutron_lbaas.extensions import lb_graph as lb_graph_ext
from neutron_lbaas.extensions import lbaas_agentschedulerv2
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.extensions import sharedpools
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import data_models
LOG = logging.getLogger(__name__)
CERT_MANAGER_PLUGIN = neutron_lbaas.common.cert_manager.get_backend()


def add_provider_configuration(type_manager, service_type):
    type_manager.add_provider_configuration(
        service_type,
        pconf.ProviderConfiguration('neutron_lbaas'))


class LoadBalancerPluginv2(loadbalancerv2.LoadBalancerPluginBaseV2,
                           agentschedulers_db.AgentSchedulerDbMixin):
    """Implementation of the Neutron Loadbalancer Service Plugin.

    This class manages the workflow of LBaaS request/response.
    Most DB related works are implemented in class
    loadbalancer_db.LoadBalancerPluginDb.
    """
    supported_extension_aliases = ["lbaasv2",
                                   "shared_pools",
                                   "l7",
                                   "lbaas_agent_schedulerv2",
                                   "service-type",
                                   "lb-graph",
                                   "lb_network_vip",
                                   "hm_max_retries_down"]
    path_prefix = loadbalancerv2.LOADBALANCERV2_PREFIX

    agent_notifiers = (
        agent_scheduler_v2.LbaasAgentSchedulerDbMixin.agent_notifiers)

    def __init__(self):
        """Initialization for the loadbalancer service plugin."""
        LOG.warning('neutron-lbaas is now deprecated. See: '
                    'https://wiki.openstack.org/wiki/Neutron/LBaaS/'
                    'Deprecation')
        self.db = ldbv2.LoadBalancerPluginDbv2()
        self.service_type_manager = st_db.ServiceTypeManager.get_instance()
        add_provider_configuration(
            self.service_type_manager, constants.LOADBALANCERV2)
        self._load_drivers()
        self.start_periodic_jobs()
        self.start_rpc_listeners()
        self.db.subscribe()
        rpc_worker = service.RpcWorker([self], worker_process_count=0)
        self.add_worker(rpc_worker)

    def start_periodic_jobs(self):
        for driver_name, driver_class in self.drivers.items():
            if hasattr(driver_class, 'get_periodic_jobs'):
                for job in self.drivers[driver_name].get_periodic_jobs():
                    self.add_agent_status_check_worker(job)

    def start_rpc_listeners(self):
        listeners = []
        for driver in self.drivers.values():
            if hasattr(driver, 'start_rpc_listeners'):
                listener = driver.start_rpc_listeners()
                listeners.append(listener)
        return listeners

    def _load_drivers(self):
        """Loads plugin-drivers specified in configuration."""
        self.drivers, self.default_provider = service_base.load_drivers(
            constants.LOADBALANCERV2, self)

        ctx = ncontext.get_admin_context()
        # stop service in case provider was removed, but resources were not
        self._check_orphan_loadbalancer_associations(ctx, self.drivers.keys())

    def _check_orphan_loadbalancer_associations(self, context, provider_names):
        """Checks remaining associations between loadbalancers and providers.

        If admin has not undeployed resources with provider that was deleted
        from configuration, neutron service is stopped. Admin must delete
        resources prior to removing providers from configuration.
        """
        used_provider_names = (
            self.db.get_provider_names_used_in_loadbalancers(context))
        lost_providers = set(
            [name for name in used_provider_names
                if name not in provider_names])
        # resources are left without provider - stop the service
        if lost_providers:
            msg = ("Delete associated load balancers before "
                   "removing providers %s") % list(lost_providers)
            LOG.error(msg)
            raise SystemExit(1)

    def _get_driver_for_provider(self, provider):
        try:
            return self.drivers[provider]
        except KeyError:
            # raise if not associated (should never be reached)
            raise n_exc.Invalid(_("Error retrieving driver for provider %s") %
                                provider)

    def _get_driver_for_loadbalancer(self, context, loadbalancer_id):
        lb = self.db.get_loadbalancer(context, loadbalancer_id)
        try:
            return self.drivers[lb.provider.provider_name]
        except KeyError:
            raise n_exc.Invalid(
                _("Error retrieving provider for load balancer. Possible "
                 "providers are %s.") % self.drivers.keys()
            )

    def _get_provider_name(self, entity):
        if ('provider' in entity and
                entity['provider'] != n_constants.ATTR_NOT_SPECIFIED):
            provider_name = pconf.normalize_provider_name(entity['provider'])
            del entity['provider']
            self.validate_provider(provider_name)
            return provider_name
        else:
            if not self.default_provider:
                raise pconf.DefaultServiceProviderNotFound(
                    service_type=constants.LOADBALANCERV2)
            if entity.get('provider'):
                del entity['provider']
            return self.default_provider

    def _call_driver_operation(self, context, driver_method, db_entity,
                               old_db_entity=None, **kwargs):
        manager_method = "%s.%s" % (driver_method.__self__.__class__.__name__,
                                    driver_method.__name__)
        LOG.info("Calling driver operation %s" % manager_method)
        try:
            if old_db_entity:
                driver_method(context, old_db_entity, db_entity, **kwargs)
            else:
                driver_method(context, db_entity, **kwargs)
        # catching and reraising agent issues
        except (lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                lbaas_agentschedulerv2.NoActiveLbaasAgent) as no_agent:
            raise no_agent
        # Pass these exceptions through to neutron
        except (exceptions.ConflictException,
                exceptions.NotFoundException,
                exceptions.NotAuthorizedException,
                exceptions.BadRequestException,
                exceptions.ServiceUnavailableException):
            raise
        except Exception as e:
            LOG.exception("There was an error in the driver")
            self._handle_driver_error(context, db_entity)
            raise loadbalancerv2.DriverError(msg=e)

    def _handle_driver_error(self, context, db_entity):
        lb_id = db_entity.root_loadbalancer.id
        self.db.update_status(context, models.LoadBalancer, lb_id,
                              n_constants.ERROR)

    def _validate_session_persistence_info(self, sp_info):
        """Performs sanity check on session persistence info.

        :param sp_info: Session persistence info
        """
        if not sp_info:
            return
        if sp_info['type'] == lb_const.SESSION_PERSISTENCE_APP_COOKIE:
            if not sp_info.get('cookie_name'):
                raise loadbalancerv2.SessionPersistenceConfigurationInvalid(
                    msg="'cookie_name' should be specified for %s"
                        " session persistence." % sp_info['type'])
        else:
            if 'cookie_name' in sp_info:
                raise loadbalancerv2.SessionPersistenceConfigurationInvalid(
                    msg="'cookie_name' is not allowed for %s"
                        " session persistence" % sp_info['type'])

    def get_plugin_type(self):
        return constants.LOADBALANCERV2

    def get_plugin_description(self):
        return "Neutron LoadBalancer Service Plugin v2"

    def _insert_provider_name_from_flavor(self, context, loadbalancer):
        """Select provider based on flavor."""

        # TODO(jwarendt) Support passing flavor metainfo from the
        # selected flavor profile into the provider, not just selecting
        # the provider, when flavor templating arrives.

        if ('provider' in loadbalancer and
            loadbalancer['provider'] != n_constants.ATTR_NOT_SPECIFIED):
            raise loadbalancerv2.ProviderFlavorConflict()

        plugin = directory.get_plugin(constants.FLAVORS)
        if not plugin:
            raise loadbalancerv2.FlavorsPluginNotLoaded()

        # Will raise FlavorNotFound if doesn't exist
        fl_db = flavors_plugin.FlavorsPlugin.get_flavor(
            plugin,
            context,
            loadbalancer['flavor_id'])

        if fl_db['service_type'] != constants.LOADBALANCERV2:
            raise lib_exc.InvalidServiceType(
                service_type=fl_db['service_type'])

        if not fl_db['enabled']:
            raise flav_exc.FlavorDisabled()

        providers = flavors_plugin.FlavorsPlugin.get_flavor_next_provider(
            plugin,
            context,
            fl_db['id'])

        provider = providers[0].get('provider')

        LOG.debug("Selected provider %s" % provider)

        loadbalancer['provider'] = provider

    def _get_tweaked_resource_attribute_map(self):
        memo = {id(n_constants.ATTR_NOT_SPECIFIED):
                n_constants.ATTR_NOT_SPECIFIED}
        ram = copy.deepcopy(attributes.RESOURCES, memo=memo)
        del ram['listeners']['loadbalancer_id']
        del ram['pools']['listener_id']
        del ram['healthmonitors']['pool_id']
        for resource in ram:
            if resource in lb_graph_ext.EXISTING_ATTR_GRAPH_ATTR_MAP:
                ram[resource].update(
                    lb_graph_ext.EXISTING_ATTR_GRAPH_ATTR_MAP[resource])
        return ram

    def _prepare_loadbalancer_graph(self, context, loadbalancer):
        """Prepares the entire user requested body of a load balancer graph

        To minimize code duplication, this method reuses the neutron API
        controller's method to do all the validation, conversion, and
        defaulting of each resource.  This reuses the RESOURCE_ATTRIBUTE_MAP
        and SUB_RESOURCE_ATTRIBUTE_MAP from the extension to enable this.
        """
        # NOTE(blogan): it is assumed the loadbalancer attributes have already
        # passed through the prepare_request_body method by nature of the
        # normal neutron wsgi workflow.  So we start with listeners since
        # that probably has not passed through the neutron wsgi workflow.
        ram = self._get_tweaked_resource_attribute_map()
        # NOTE(blogan): members are not populated in the attributes.RAM so
        # our only option is to use the original extension definition of member
        # to validate.  If members ever need something added to it then it too
        # will need to be added here.
        prepped_lb = napi_base.Controller.prepare_request_body(
            context, {'loadbalancer': loadbalancer}, True, 'loadbalancer',
            ram['loadbalancers']
        )
        sub_ram = loadbalancerv2.SUB_RESOURCE_ATTRIBUTE_MAP
        sub_ram.update(l7.SUB_RESOURCE_ATTRIBUTE_MAP)
        prepped_listeners = []
        for listener in loadbalancer.get('listeners', []):
            prepped_listener = napi_base.Controller.prepare_request_body(
                context, {'listener': listener}, True, 'listener',
                ram['listeners'])
            l7policies = listener.get('l7policies')
            if l7policies and l7policies != n_constants.ATTR_NOT_SPECIFIED:
                prepped_policies = []
                for policy in l7policies:
                    prepped_policy = napi_base.Controller.prepare_request_body(
                        context, {'l7policy': policy}, True, 'l7policy',
                        ram['l7policies'])
                    l7rules = policy.get('rules')
                    redirect_pool = policy.get('redirect_pool')
                    if l7rules and l7rules != n_constants.ATTR_NOT_SPECIFIED:
                        prepped_rules = []
                        for rule in l7rules:
                            prepped_rule = (
                                napi_base.Controller.prepare_request_body(
                                    context, {'l7rule': rule}, True, 'l7rule',
                                    sub_ram['rules']['parameters']))
                            prepped_rules.append(prepped_rule)
                        prepped_policy['l7_rules'] = prepped_rules
                    if (redirect_pool and
                            redirect_pool != n_constants.ATTR_NOT_SPECIFIED):
                        prepped_r_pool = (
                            napi_base.Controller.prepare_request_body(
                                context, {'pool': redirect_pool}, True, 'pool',
                                ram['pools']))
                        prepped_r_members = []
                        for member in redirect_pool.get('members', []):
                            prepped_r_member = (
                                napi_base.Controller.prepare_request_body(
                                    context, {'member': member},
                                    True, 'member',
                                    sub_ram['members']['parameters']))
                            prepped_r_members.append(prepped_r_member)
                        prepped_r_pool['members'] = prepped_r_members
                        r_hm = redirect_pool.get('healthmonitor')
                        if r_hm and r_hm != n_constants.ATTR_NOT_SPECIFIED:
                            prepped_r_hm = (
                                napi_base.Controller.prepare_request_body(
                                    context, {'healthmonitor': r_hm},
                                    True, 'healthmonitor',
                                    ram['healthmonitors']))
                            prepped_r_pool['healthmonitor'] = prepped_r_hm
                        prepped_policy['redirect_pool'] = redirect_pool
                    prepped_policies.append(prepped_policy)
                prepped_listener['l7_policies'] = prepped_policies
            pool = listener.get('default_pool')
            if pool and pool != n_constants.ATTR_NOT_SPECIFIED:
                prepped_pool = napi_base.Controller.prepare_request_body(
                    context, {'pool': pool}, True, 'pool',
                    ram['pools'])
                prepped_members = []
                for member in pool.get('members', []):
                    prepped_member = napi_base.Controller.prepare_request_body(
                        context, {'member': member}, True, 'member',
                        sub_ram['members']['parameters'])
                    prepped_members.append(prepped_member)
                prepped_pool['members'] = prepped_members
                hm = pool.get('healthmonitor')
                if hm and hm != n_constants.ATTR_NOT_SPECIFIED:
                    prepped_hm = napi_base.Controller.prepare_request_body(
                        context, {'healthmonitor': hm}, True, 'healthmonitor',
                        ram['healthmonitors'])
                    prepped_pool['healthmonitor'] = prepped_hm
                prepped_listener['default_pool'] = prepped_pool
            prepped_listeners.append(prepped_listener)
        prepped_lb['listeners'] = prepped_listeners
        return loadbalancer

    def create_loadbalancer(self, context, loadbalancer):
        loadbalancer = loadbalancer.get('loadbalancer')
        if loadbalancer['flavor_id'] != n_constants.ATTR_NOT_SPECIFIED:
            self._insert_provider_name_from_flavor(context, loadbalancer)
        else:
            del loadbalancer['flavor_id']
        provider_name = self._get_provider_name(loadbalancer)
        driver = self.drivers[provider_name]
        lb_db = self.db.create_loadbalancer(
            context, loadbalancer,
            allocate_vip=not driver.load_balancer.allocates_vip)
        self.service_type_manager.add_resource_association(
            context,
            constants.LOADBALANCERV2,
            provider_name, lb_db.id)
        create_method = (driver.load_balancer.create_and_allocate_vip
                         if driver.load_balancer.allocates_vip
                         else driver.load_balancer.create)
        try:
            self._call_driver_operation(context, create_method, lb_db)
        except (lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                lbaas_agentschedulerv2.NoActiveLbaasAgent) as no_agent:
            self.db.delete_loadbalancer(context, lb_db.id)
            raise no_agent
        return self.db.get_loadbalancer(context, lb_db.id).to_api_dict()

    def create_graph(self, context, graph):
        loadbalancer = graph.get('graph', {}).get('loadbalancer')
        loadbalancer = self._prepare_loadbalancer_graph(context, loadbalancer)
        if loadbalancer['flavor_id'] != n_constants.ATTR_NOT_SPECIFIED:
            self._insert_provider_name_from_flavor(context, loadbalancer)
        else:
            del loadbalancer['flavor_id']
        provider_name = self._get_provider_name(loadbalancer)
        driver = self.drivers[provider_name]
        if not driver.load_balancer.allows_create_graph:
            raise lb_graph_ext.ProviderCannotCreateLoadBalancerGraph
        lb_db = self.db.create_loadbalancer_graph(
            context, loadbalancer,
            allocate_vip=not driver.load_balancer.allocates_vip)
        self.service_type_manager.add_resource_association(
            context, constants.LOADBALANCERV2, provider_name, lb_db.id)
        create_method = (driver.load_balancer.create_and_allocate_vip
                         if driver.load_balancer.allocates_vip
                         else driver.load_balancer.create)
        self._call_driver_operation(context, create_method, lb_db)
        api_lb = {'loadbalancer': self.db.get_loadbalancer(
            context, lb_db.id).to_api_dict(full_graph=True)}
        return api_lb

    def update_loadbalancer(self, context, id, loadbalancer):
        loadbalancer = loadbalancer.get('loadbalancer')
        old_lb = self.db.get_loadbalancer(context, id)
        self.db.test_and_set_status(context, models.LoadBalancer, id,
                                    n_constants.PENDING_UPDATE)
        try:
            updated_lb = self.db.update_loadbalancer(
                context, id, loadbalancer)
        except Exception as exc:
            self.db.update_status(context, models.LoadBalancer, id,
                                  old_lb.provisioning_status)
            raise exc
        driver = self._get_driver_for_provider(old_lb.provider.provider_name)
        self._call_driver_operation(context,
                                    driver.load_balancer.update,
                                    updated_lb, old_db_entity=old_lb)
        return self.db.get_loadbalancer(context, id).to_api_dict()

    def delete_loadbalancer(self, context, id):
        old_lb = self.db.get_loadbalancer(context, id)
        if old_lb.listeners:
            raise loadbalancerv2.EntityInUse(
                entity_using=models.LoadBalancer.NAME,
                id=old_lb.listeners[0].id,
                entity_in_use=models.Listener.NAME)
        if old_lb.pools:
            raise loadbalancerv2.EntityInUse(
                entity_using=models.LoadBalancer.NAME,
                id=old_lb.pools[0].id,
                entity_in_use=models.PoolV2.NAME)
        self.db.test_and_set_status(context, models.LoadBalancer, id,
                                    n_constants.PENDING_DELETE)
        driver = self._get_driver_for_provider(old_lb.provider.provider_name)
        db_lb = self.db.get_loadbalancer(context, id)
        self._call_driver_operation(
            context, driver.load_balancer.delete, db_lb)

    def get_loadbalancer(self, context, id, fields=None):
        return self.db.get_loadbalancer(context, id).to_api_dict()

    def get_loadbalancers(self, context, filters=None, fields=None):
        return self.db.get_loadbalancers_as_api_dict(context, filters=filters)

    def _validate_tls(self, listener, curr_listener=None):
        def validate_tls_container(container_ref):
            cert_mgr = CERT_MANAGER_PLUGIN.CertManager()

            if curr_listener:
                lb_id = curr_listener['loadbalancer_id']
                tenant_id = curr_listener['tenant_id']
            else:
                lb_id = listener.get('loadbalancer_id')
                tenant_id = listener.get('tenant_id')

            try:
                cert_container = cert_mgr.get_cert(
                    project_id=tenant_id,
                    cert_ref=container_ref,
                    resource_ref=cert_mgr.get_service_url(lb_id))
            except Exception as e:
                if hasattr(e, 'status_code') and e.status_code == 404:
                    raise loadbalancerv2.TLSContainerNotFound(
                        container_id=container_ref)
                else:
                    # Could be a keystone configuration error...
                    err_msg = encodeutils.exception_to_unicode(e)
                    raise loadbalancerv2.CertManagerError(
                        ref=container_ref, reason=err_msg
                    )

            try:
                cert_parser.validate_cert(
                    cert_container.get_certificate(),
                    private_key=cert_container.get_private_key(),
                    private_key_passphrase=(
                        cert_container.get_private_key_passphrase()),
                    intermediates=cert_container.get_intermediates())
            except Exception as e:
                try:
                    cert_mgr.delete_cert(
                        project_id=tenant_id,
                        cert_ref=container_ref,
                        resource_ref=cert_mgr.get_service_url(lb_id))
                except Exception:
                    LOG.error('Unable to delete cert', exc_info=True)
                raise loadbalancerv2.TLSContainerInvalid(
                    container_id=container_ref, reason=str(e))

        def validate_tls_containers(to_validate):
            for container_ref in to_validate:
                validate_tls_container(container_ref)

        to_validate = []
        if not curr_listener:
            if not listener['default_tls_container_ref']:
                raise loadbalancerv2.TLSDefaultContainerNotSpecified()
            to_validate.extend([listener['default_tls_container_ref']])
            if 'sni_container_refs' in listener:
                to_validate.extend(listener['sni_container_refs'])
        elif curr_listener['provisioning_status'] == n_constants.ERROR:
            to_validate.extend(curr_listener['default_tls_container_id'])
            to_validate.extend([
                container['tls_container_id'] for container in (
                    curr_listener['sni_containers'])])
        else:
            if (curr_listener['default_tls_container_id'] !=
                    listener['default_tls_container_ref']):
                to_validate.extend([listener['default_tls_container_ref']])

            if ('sni_container_refs' in listener and
                    [container['tls_container_id'] for container in (
                        curr_listener['sni_containers'])] !=
                    listener['sni_container_refs']):
                to_validate.extend(listener['sni_container_refs'])

        if len(to_validate) > 0:
            validate_tls_containers(to_validate)

        return len(to_validate) > 0

    def _check_pool_loadbalancer_match(self, context, pool_id, lb_id):
        lb = self.db.get_loadbalancer(context, lb_id)
        pool = self.db.get_pool(context, pool_id)
        if not lb.id == pool.loadbalancer.id:
            raise sharedpools.ListenerAndPoolMustBeOnSameLoadbalancer()

    def create_listener(self, context, listener):
        listener = listener.get('listener')
        lb_id = listener.get('loadbalancer_id')
        default_pool_id = listener.get('default_pool_id')
        if default_pool_id:
            self._check_pool_exists(context, default_pool_id)
            # Get the loadbalancer from the default_pool_id
            if not lb_id:
                default_pool = self.db.get_pool(context, default_pool_id)
                lb_id = default_pool.loadbalancer.id
                listener['loadbalancer_id'] = lb_id
            else:
                self._check_pool_loadbalancer_match(
                    context, default_pool_id, lb_id)
        elif not lb_id:
            raise sharedpools.ListenerMustHaveLoadbalancer()
        self.db.test_and_set_status(context, models.LoadBalancer, lb_id,
                                    n_constants.PENDING_UPDATE)

        try:
            if listener['protocol'] == lb_const.PROTOCOL_TERMINATED_HTTPS:
                self._validate_tls(listener)
            listener_db = self.db.create_listener(context, listener)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, lb_id)
            raise exc
        driver = self._get_driver_for_loadbalancer(
            context, listener_db.loadbalancer_id)
        self._call_driver_operation(
            context, driver.listener.create, listener_db)

        return self.db.get_listener(context, listener_db.id).to_api_dict()

    def _check_listener_pool_lb_match(self, context, listener_id, pool_id):
        listener = self.db.get_listener(context, listener_id)
        pool = self.db.get_pool(context, pool_id)
        if not listener.loadbalancer.id == pool.loadbalancer.id:
            raise sharedpools.ListenerAndPoolMustBeOnSameLoadbalancer()

    def update_listener(self, context, id, listener):
        listener = listener.get('listener')
        curr_listener_db = self.db.get_listener(context, id)
        default_pool_id = listener.get('default_pool_id')
        if default_pool_id:
            self._check_listener_pool_lb_match(
                context, id, default_pool_id)
        self.db.test_and_set_status(context, models.Listener, id,
                                    n_constants.PENDING_UPDATE)
        try:
            curr_listener = curr_listener_db.to_dict()

            if 'default_tls_container_ref' not in listener:
                listener['default_tls_container_ref'] = (
                    # NOTE(blogan): not changing to ref bc this dictionary is
                    # created from a data model
                    curr_listener['default_tls_container_id'])
            if 'sni_container_refs' not in listener:
                listener['sni_container_ids'] = [
                    container['tls_container_id'] for container in (
                        curr_listener['sni_containers'])]

            tls_containers_changed = False
            if curr_listener['protocol'] == lb_const.PROTOCOL_TERMINATED_HTTPS:
                tls_containers_changed = self._validate_tls(
                    listener, curr_listener=curr_listener)
            listener_db = self.db.update_listener(
                context, id, listener,
                tls_containers_changed=tls_containers_changed)
        except Exception as exc:
            self.db.update_status(
                context,
                models.LoadBalancer,
                curr_listener_db.loadbalancer.id,
                provisioning_status=n_constants.ACTIVE
            )
            self.db.update_status(
                context,
                models.Listener,
                curr_listener_db.id,
                provisioning_status=n_constants.ACTIVE
            )
            raise exc

        driver = self._get_driver_for_loadbalancer(
            context, listener_db.loadbalancer_id)
        self._call_driver_operation(
            context,
            driver.listener.update,
            listener_db,
            old_db_entity=curr_listener_db)

        return self.db.get_listener(context, id).to_api_dict()

    def delete_listener(self, context, id):
        old_listener = self.db.get_listener(context, id)
        if old_listener.l7_policies:
            raise loadbalancerv2.EntityInUse(
                entity_using=models.Listener.NAME,
                id=old_listener.l7_policies[0].id,
                entity_in_use=models.L7Policy.NAME)
        self.db.test_and_set_status(context, models.Listener, id,
                                    n_constants.PENDING_DELETE)
        listener_db = self.db.get_listener(context, id)

        driver = self._get_driver_for_loadbalancer(
            context, listener_db.loadbalancer_id)
        self._call_driver_operation(
            context, driver.listener.delete, listener_db)

    def get_listener(self, context, id, fields=None):
        return self.db.get_listener(context, id).to_api_dict()

    def get_listeners(self, context, filters=None, fields=None):
        return self.db.get_listeners_as_api_dict(
            context, filters=filters)

    def create_pool(self, context, pool):
        pool = pool.get('pool')
        listener_id = pool.get('listener_id')
        listeners = pool.get('listeners', [])
        if listener_id:
            listeners.append(listener_id)
        lb_id = pool.get('loadbalancer_id')
        db_listeners = []
        for l in listeners:
            db_l = self.db.get_listener(context, l)
            db_listeners.append(db_l)
            # Take the pool's loadbalancer_id from the first listener found
            # if it wasn't specified in the API call.
            if not lb_id:
                lb_id = db_l.loadbalancer.id
            # All specified listeners must be on the same loadbalancer
            if db_l.loadbalancer.id != lb_id:
                raise sharedpools.ListenerAndPoolMustBeOnSameLoadbalancer()
            if db_l.default_pool_id:
                raise sharedpools.ListenerDefaultPoolAlreadySet(
                    listener_id=db_l.id, pool_id=db_l.default_pool_id)
            if ((pool['protocol'], db_l.protocol)
                not in lb_const.LISTENER_POOL_COMPATIBLE_PROTOCOLS):
                raise loadbalancerv2.ListenerPoolProtocolMismatch(
                    listener_proto=db_l.protocol,
                    pool_proto=pool['protocol'])
        if not lb_id:
            raise sharedpools.PoolMustHaveLoadbalancer()
        pool['loadbalancer_id'] = lb_id
        self._validate_session_persistence_info(
            pool.get('session_persistence'))
        # SQLAlchemy gets strange ideas about populating the pool if we don't
        # blank out the listeners at this point.
        del pool['listener_id']
        pool['listeners'] = []
        self.db.test_and_set_status(context, models.LoadBalancer,
                                    lb_id, n_constants.PENDING_UPDATE)
        db_pool = self.db.create_pool(context, pool)
        for db_l in db_listeners:
            try:
                self.db.update_listener(context, db_l.id,
                                        {'default_pool_id': db_pool.id})
            except Exception as exc:
                self.db.update_loadbalancer_provisioning_status(
                    context, db_pool.loadbalancer_id)
                raise exc
        # Reload the pool from the DB to re-populate pool.listeners
        # before calling the driver
        db_pool = self.db.get_pool(context, db_pool.id)
        driver = self._get_driver_for_loadbalancer(
            context, db_pool.loadbalancer_id)
        self._call_driver_operation(context, driver.pool.create, db_pool)
        return db_pool.to_api_dict()

    def update_pool(self, context, id, pool):
        pool = pool.get('pool')
        self._validate_session_persistence_info(
            pool.get('session_persistence'))
        old_pool = self.db.get_pool(context, id)
        self.db.test_and_set_status(context, models.PoolV2, id,
                                    n_constants.PENDING_UPDATE)
        try:
            updated_pool = self.db.update_pool(context, id, pool)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, old_pool.root_loadbalancer.id)
            raise exc

        driver = self._get_driver_for_loadbalancer(
            context, updated_pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.pool.update,
                                    updated_pool,
                                    old_db_entity=old_pool)

        return self.db.get_pool(context, id).to_api_dict()

    def delete_pool(self, context, id):
        old_pool = self.db.get_pool(context, id)
        if old_pool.healthmonitor:
            raise loadbalancerv2.EntityInUse(
                entity_using=models.PoolV2.NAME,
                id=old_pool.healthmonitor.id,
                entity_in_use=models.HealthMonitorV2.NAME)
        self.db.test_and_set_status(context, models.PoolV2, id,
                                    n_constants.PENDING_DELETE)
        db_pool = self.db.get_pool(context, id)

        driver = self._get_driver_for_loadbalancer(
            context, db_pool.loadbalancer_id)
        self._call_driver_operation(context, driver.pool.delete, db_pool)

    def get_pools(self, context, filters=None, fields=None):
        return self.db.get_pools_as_api_dict(
            context, filters=filters)

    def get_pool(self, context, id, fields=None):
        return self.db.get_pool(context, id).to_api_dict()

    def _check_pool_exists(self, context, pool_id):
        if not self.db._resource_exists(context, models.PoolV2, pool_id):
            raise loadbalancerv2.EntityNotFound(name=models.PoolV2.NAME,
                                                id=pool_id)

    def create_pool_member(self, context, pool_id, member):
        member = member.get('member')
        self.db.check_subnet_exists(context, member['subnet_id'])
        db_pool = self.db.get_pool(context, pool_id)
        self.db.test_and_set_status(context, models.LoadBalancer,
                                    db_pool.root_loadbalancer.id,
                                    n_constants.PENDING_UPDATE)
        try:
            member_db = self.db.create_pool_member(context, member, pool_id)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, db_pool.root_loadbalancer.id)
            raise exc

        driver = self._get_driver_for_loadbalancer(
            context, member_db.pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.member.create,
                                    member_db)

        return self.db.get_pool_member(context, member_db.id).to_api_dict()

    def update_pool_member(self, context, id, pool_id, member):
        self._check_pool_exists(context, pool_id)
        member = member.get('member')
        old_member = self.db.get_pool_member(context, id)
        self.db.test_and_set_status(context, models.MemberV2, id,
                                    n_constants.PENDING_UPDATE)
        try:
            updated_member = self.db.update_pool_member(context, id, member)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, old_member.pool.loadbalancer.id)
            raise exc

        driver = self._get_driver_for_loadbalancer(
            context, updated_member.pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.member.update,
                                    updated_member,
                                    old_db_entity=old_member)

        return self.db.get_pool_member(context, id).to_api_dict()

    def delete_pool_member(self, context, id, pool_id):
        self._check_pool_exists(context, pool_id)
        self.db.test_and_set_status(context, models.MemberV2, id,
                                    n_constants.PENDING_DELETE)
        db_member = self.db.get_pool_member(context, id)

        driver = self._get_driver_for_loadbalancer(
            context, db_member.pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.member.delete,
                                    db_member)

    def get_pool_members(self, context, pool_id, filters=None, fields=None):
        self._check_pool_exists(context, pool_id)
        if not filters:
            filters = {}
        filters['pool_id'] = [pool_id]
        return self.db.get_pool_members_as_api_dict(
            context, filters=filters)

    def get_pool_member(self, context, id, pool_id, fields=None):
        self._check_pool_exists(context, pool_id)
        return self.db.get_pool_member(context, id).to_api_dict()

    def _check_pool_already_has_healthmonitor(self, context, pool_id):
        pool = self.db.get_pool(context, pool_id)
        if pool.healthmonitor:
            raise loadbalancerv2.OneHealthMonitorPerPool(
                pool_id=pool_id, hm_id=pool.healthmonitor.id)

    def create_healthmonitor(self, context, healthmonitor):
        healthmonitor = healthmonitor.get('healthmonitor')
        pool_id = healthmonitor.pop('pool_id')
        self._check_pool_exists(context, pool_id)
        self._check_pool_already_has_healthmonitor(context, pool_id)
        db_pool = self.db.get_pool(context, pool_id)
        self.db.test_and_set_status(context, models.LoadBalancer,
                                    db_pool.root_loadbalancer.id,
                                    n_constants.PENDING_UPDATE)
        try:
            db_hm = self.db.create_healthmonitor_on_pool(context, pool_id,
                                                         healthmonitor)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, db_pool.root_loadbalancer.id)
            raise exc
        driver = self._get_driver_for_loadbalancer(
            context, db_hm.pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.health_monitor.create,
                                    db_hm)
        return self.db.get_healthmonitor(context, db_hm.id).to_api_dict()

    def update_healthmonitor(self, context, id, healthmonitor):
        healthmonitor = healthmonitor.get('healthmonitor')
        old_hm = self.db.get_healthmonitor(context, id)
        self.db.test_and_set_status(context, models.HealthMonitorV2, id,
                                    n_constants.PENDING_UPDATE)
        try:
            updated_hm = self.db.update_healthmonitor(context, id,
                                                      healthmonitor)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, old_hm.root_loadbalancer.id)
            raise exc

        driver = self._get_driver_for_loadbalancer(
            context, updated_hm.pool.loadbalancer_id)
        self._call_driver_operation(context,
                                    driver.health_monitor.update,
                                    updated_hm,
                                    old_db_entity=old_hm)

        return self.db.get_healthmonitor(context, updated_hm.id).to_api_dict()

    def delete_healthmonitor(self, context, id):
        self.db.test_and_set_status(context, models.HealthMonitorV2, id,
                                    n_constants.PENDING_DELETE)
        db_hm = self.db.get_healthmonitor(context, id)

        driver = self._get_driver_for_loadbalancer(
            context, db_hm.pool.loadbalancer_id)
        self._call_driver_operation(
            context, driver.health_monitor.delete, db_hm)

    def get_healthmonitor(self, context, id, fields=None):
        return self.db.get_healthmonitor(context, id).to_api_dict()

    def get_healthmonitors(self, context, filters=None, fields=None):
        return self.db.get_healthmonitors_as_api_dict(
            context, filters=filters)

    def stats(self, context, loadbalancer_id):
        lb = self.db.get_loadbalancer(context, loadbalancer_id)
        driver = self._get_driver_for_loadbalancer(context, loadbalancer_id)
        stats_data = driver.load_balancer.stats(context, lb)
        # if we get something from the driver -
        # update the db and return the value from db
        # else - return what we have in db
        if stats_data:
            self.db.update_loadbalancer_stats(context, loadbalancer_id,
                                              stats_data)
        db_stats = self.db.stats(context, loadbalancer_id)
        return {'stats': db_stats.to_api_dict()}

    def create_l7policy(self, context, l7policy):
        l7policy = l7policy.get('l7policy')
        l7policy_db = self.db.create_l7policy(context, l7policy)

        if l7policy_db.attached_to_loadbalancer():
            driver = self._get_driver_for_loadbalancer(
                context, l7policy_db.listener.loadbalancer_id)
            self._call_driver_operation(context,
                                        driver.l7policy.create,
                                        l7policy_db)

        return l7policy_db.to_dict()

    def update_l7policy(self, context, id, l7policy):
        l7policy = l7policy.get('l7policy')
        old_l7policy = self.db.get_l7policy(context, id)
        self.db.test_and_set_status(context, models.L7Policy, id,
                                    n_constants.PENDING_UPDATE)
        try:
            updated_l7policy = self.db.update_l7policy(
                context, id, l7policy)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, old_l7policy.root_loadbalancer.id)
            raise exc

        if (updated_l7policy.attached_to_loadbalancer() or
                old_l7policy.attached_to_loadbalancer()):
            if updated_l7policy.attached_to_loadbalancer():
                driver = self._get_driver_for_loadbalancer(
                    context, updated_l7policy.listener.loadbalancer_id)
            else:
                driver = self._get_driver_for_loadbalancer(
                    context, old_l7policy.listener.loadbalancer_id)
            self._call_driver_operation(context,
                                        driver.l7policy.update,
                                        updated_l7policy,
                                        old_db_entity=old_l7policy)

        return self.db.get_l7policy(context, updated_l7policy.id).to_api_dict()

    def delete_l7policy(self, context, id):
        self.db.test_and_set_status(context, models.L7Policy, id,
                                    n_constants.PENDING_DELETE)
        l7policy_db = self.db.get_l7policy(context, id)

        if l7policy_db.attached_to_loadbalancer():
            driver = self._get_driver_for_loadbalancer(
                context, l7policy_db.listener.loadbalancer_id)
            self._call_driver_operation(context, driver.l7policy.delete,
                                        l7policy_db)
        else:
            self.db.delete_l7policy(context, id)

    def get_l7policies(self, context, filters=None, fields=None):
        return self.db.get_l7policies_as_api_dict(
            context, filters=filters)

    def get_l7policy(self, context, id, fields=None):
        return self.db.get_l7policy(context, id).to_api_dict()

    def _check_l7policy_exists(self, context, l7policy_id):
        if not self.db._resource_exists(context, models.L7Policy, l7policy_id):
            raise loadbalancerv2.EntityNotFound(name=models.L7Policy.NAME,
                                                id=l7policy_id)

    def create_l7policy_rule(self, context, rule, l7policy_id):
        rule = rule.get('rule')
        rule_db = self.db.create_l7policy_rule(context, rule, l7policy_id)

        if rule_db.attached_to_loadbalancer():
            driver = self._get_driver_for_loadbalancer(
                context, rule_db.policy.listener.loadbalancer_id)
            self._call_driver_operation(context,
                                        driver.l7rule.create,
                                        rule_db)
        else:
            self.db.update_status(context, models.L7Rule, rule_db.id,
                                  lb_const.DEFERRED)

        return rule_db.to_dict()

    def update_l7policy_rule(self, context, id, rule, l7policy_id):
        rule = rule.get('rule')
        old_rule_db = self.db.get_l7policy_rule(context, id, l7policy_id)
        self.db.test_and_set_status(context, models.L7Rule, id,
                                    n_constants.PENDING_UPDATE)
        try:
            upd_rule_db = self.db.update_l7policy_rule(
                context, id, rule, l7policy_id)
        except Exception as exc:
            self.db.update_loadbalancer_provisioning_status(
                context, old_rule_db.root_loadbalancer.id)
            raise exc

        if (upd_rule_db.attached_to_loadbalancer() or
                old_rule_db.attached_to_loadbalancer()):
            if upd_rule_db.attached_to_loadbalancer():
                driver = self._get_driver_for_loadbalancer(
                    context, upd_rule_db.policy.listener.loadbalancer_id)
            else:
                driver = self._get_driver_for_loadbalancer(
                    context, old_rule_db.policy.listener.loadbalancer_id)
            self._call_driver_operation(context,
                                        driver.l7rule.update,
                                        upd_rule_db,
                                        old_db_entity=old_rule_db)
        else:
            self.db.update_status(context, models.L7Rule, id,
                                  lb_const.DEFERRED)

        return upd_rule_db.to_dict()

    def delete_l7policy_rule(self, context, id, l7policy_id):
        self.db.test_and_set_status(context, models.L7Rule, id,
                                    n_constants.PENDING_DELETE)
        rule_db = self.db.get_l7policy_rule(context, id, l7policy_id)

        if rule_db.attached_to_loadbalancer():
            driver = self._get_driver_for_loadbalancer(
                context, rule_db.policy.listener.loadbalancer_id)
            self._call_driver_operation(context, driver.l7rule.delete,
                                        rule_db)
        else:
            self.db.delete_l7policy_rule(context, id, l7policy_id)

    def get_l7policy_rules(self, context, l7policy_id,
                           filters=None, fields=None):
        self._check_l7policy_exists(context, l7policy_id)
        return self.db.get_l7policy_rules_as_api_dict(
            context, l7policy_id, filters=filters)

    def get_l7policy_rule(self, context, id, l7policy_id, fields=None):
        self._check_l7policy_exists(context, l7policy_id)
        return self.db.get_l7policy_rule(
            context, id, l7policy_id).to_api_dict()

    def validate_provider(self, provider):
        if provider not in self.drivers:
            raise pconf.ServiceProviderNotFound(
                provider=provider, service_type=constants.LOADBALANCERV2)

    def _default_status(self, obj, exclude=None, **kw):
        exclude = exclude or []
        status = {}
        status["id"] = obj.id
        if "provisioning_status" not in exclude:
            status["provisioning_status"] = obj.provisioning_status
        if "operating_status" not in exclude:
            status["operating_status"] = obj.operating_status
        for key, value in kw.items():
            status[key] = value
        try:
            status['name'] = getattr(obj, 'name')
        except AttributeError:
            pass
        return status

    def _disable_entity_and_children(self, obj):
        DISABLED = lb_const.DISABLED
        d = {}
        if isinstance(obj, data_models.LoadBalancer):
            d = {'loadbalancer': {'id': obj.id, 'operating_status': DISABLED,
                'provisioning_status': obj.provisioning_status,
                'name': obj.name, 'listeners': []}}
            for listener in obj.listeners:
                listener_dict = self._disable_entity_and_children(listener)
                d['loadbalancer']['listeners'].append(listener_dict)
        if isinstance(obj, data_models.Listener):
            d = {'id': obj.id, 'operating_status': DISABLED,
                 'provisioning_status': obj.provisioning_status,
                 'name': obj.name, 'pools': [], 'l7policies': []}
            if obj.default_pool:
                pool_dict = self._disable_entity_and_children(obj.default_pool)
                d['pools'].append(pool_dict)
            for policy in obj.l7_policies:
                policy_dict = self._disable_entity_and_children(policy)
                d['l7policies'].append(policy_dict)
        if isinstance(obj, data_models.L7Policy):
            d = {'id': obj.id,
                 'provisioning_status': obj.provisioning_status,
                 'name': obj.name, 'rules': []}
            for rule in obj.rules:
                rule_dict = self._disable_entity_and_children(rule)
                d['rules'].append(rule_dict)
        if isinstance(obj, data_models.L7Rule):
            d = {'id': obj.id,
                 'provisioning_status': obj.provisioning_status,
                 'type': obj.type}
        if isinstance(obj, data_models.Pool):
            d = {'id': obj.id, 'operating_status': DISABLED,
                 'provisioning_status': obj.provisioning_status,
                 'name': obj.name, 'members': [], 'healthmonitor': {}}
            for member in obj.members:
                member_dict = self._disable_entity_and_children(member)
                d['members'].append(member_dict)
            d['healthmonitor'] = self._disable_entity_and_children(
                obj.healthmonitor)
        if isinstance(obj, data_models.HealthMonitor):
            d = {'id': obj.id, 'provisioning_status': obj.provisioning_status,
                 'type': obj.type}
        if isinstance(obj, data_models.Member):
            d = {'id': obj.id, 'operating_status': DISABLED,
                 'provisioning_status': obj.provisioning_status,
                 'address': obj.address, 'protocol_port': obj.protocol_port}
        return d

    def statuses(self, context, loadbalancer_id):
        OS = "operating_status"
        lb = self.db.get_loadbalancer(context, loadbalancer_id)
        if not lb.admin_state_up:
            return {"statuses": self._disable_entity_and_children(lb)}
        lb_status = self._default_status(lb, listeners=[], pools=[])
        statuses = {"statuses": {"loadbalancer": lb_status}}
        if self._is_degraded(lb):
            self._set_degraded(lb_status)
        for curr_listener in lb.listeners:
            if not curr_listener.admin_state_up:
                lb_status["listeners"].append(
                    self._disable_entity_and_children(curr_listener)
                )
                continue
            listener_status = self._default_status(curr_listener,
                                                   pools=[], l7policies=[])
            lb_status["listeners"].append(listener_status)
            if self._is_degraded(curr_listener):
                self._set_degraded(lb_status)

            for policy in curr_listener.l7_policies:
                if not policy.admin_state_up:
                    listener_status["l7policies"].append(
                        self._disable_entity_and_children(policy))
                    continue
                policy_opts = {"action": policy.action, "rules": []}
                policy_status = self._default_status(policy, exclude=[OS],
                                                     **policy_opts)
                listener_status["l7policies"].append(policy_status)
                if self._is_degraded(policy, exclude=[OS]):
                    self._set_degraded(policy_status, listener_status,
                                       lb_status)
                for rule in policy.rules:
                    if not rule.admin_state_up:
                        policy_status["rules"].append(
                            self._disable_entity_and_children(rule))
                        continue
                    rule_opts = {"type": rule.type}
                    rule_status = self._default_status(rule, exclude=[OS],
                                                       **rule_opts)
                    policy_status["rules"].append(rule_status)
                    if self._is_degraded(rule, exclude=[OS]):
                        self._set_degraded(rule_status, policy_status,
                                           listener_status,
                                           lb_status)
            if not curr_listener.default_pool:
                continue
            if not curr_listener.default_pool.admin_state_up:
                listener_status["pools"].append(
                    self._disable_entity_and_children(
                        curr_listener.default_pool))
                continue
            pool_status = self._default_status(curr_listener.default_pool,
                                              members=[], healthmonitor={})
            listener_status["pools"].append(pool_status)
            if (pool_status["id"] not in
                [ps["id"] for ps in lb_status["pools"]]):
                lb_status["pools"].append(pool_status)
            if self._is_degraded(curr_listener.default_pool):
                self._set_degraded(listener_status, lb_status)
            members = curr_listener.default_pool.members
            for curr_member in members:
                if not curr_member.admin_state_up:
                    pool_status["members"].append(
                        self._disable_entity_and_children(curr_member))
                    continue
                member_opts = {"address": curr_member.address,
                               "protocol_port": curr_member.protocol_port}
                member_status = self._default_status(curr_member,
                                                     **member_opts)
                pool_status["members"].append(member_status)
                if self._is_degraded(curr_member):
                    self._set_degraded(pool_status, listener_status,
                                       lb_status)
            healthmonitor = curr_listener.default_pool.healthmonitor
            if healthmonitor:
                if not healthmonitor.admin_state_up:
                    dhm = self._disable_entity_and_children(healthmonitor)
                    hm_status = dhm
                else:
                    hm_status = self._default_status(healthmonitor,
                                exclude=[OS], type=healthmonitor.type)
                    if self._is_degraded(healthmonitor, exclude=[OS]):
                        self._set_degraded(pool_status, listener_status,
                                           lb_status)
            else:
                hm_status = {}
            pool_status["healthmonitor"] = hm_status

        # Needed for pools not associated with a listener
        for curr_pool in lb.pools:
            if curr_pool.id in [ps["id"] for ps in lb_status["pools"]]:
                continue
            if not curr_pool.admin_state_up:
                lb_status["pools"].append(
                    self._disable_entity_and_children(curr_pool))
                continue
            pool_status = self._default_status(curr_pool, members=[],
                                               healthmonitor={})
            lb_status["pools"].append(pool_status)
            if self._is_degraded(curr_pool):
                self._set_degraded(lb_status)
            members = curr_pool.members
            for curr_member in members:
                if not curr_member.admin_state_up:
                    pool_status["members"].append(
                        self._disable_entity_and_children(curr_member))
                    continue
                member_opts = {"address": curr_member.address,
                               "protocol_port": curr_member.protocol_port}
                member_status = self._default_status(curr_member,
                                                     **member_opts)
                pool_status["members"].append(member_status)
                if self._is_degraded(curr_member):
                    self._set_degraded(pool_status, lb_status)
            healthmonitor = curr_pool.healthmonitor
            if healthmonitor:
                if not healthmonitor.admin_state_up:
                    dhm = self._disable_entity_and_children(healthmonitor)
                    hm_status = dhm
                else:
                    hm_status = self._default_status(healthmonitor,
                                exclude=[OS], type=healthmonitor.type)
                    if self._is_degraded(healthmonitor, exclude=[OS]):
                        self._set_degraded(pool_status, listener_status,
                                           lb_status)
            else:
                hm_status = {}
            pool_status["healthmonitor"] = hm_status
        return statuses

    def _set_degraded(self, *objects):
        for obj in objects:
            obj["operating_status"] = lb_const.DEGRADED

    def _is_degraded(self, obj, exclude=None):
        exclude = exclude or []
        if "provisioning_status" not in exclude:
            if obj.provisioning_status == n_constants.ERROR:
                return True
        if "operating_status" not in exclude:
            if ((obj.operating_status != lb_const.ONLINE) and
                (obj.operating_status != lb_const.NO_MONITOR)):
                return True
        return False

    # NOTE(brandon-logan): these need to be concrete methods because the
    # neutron request pipeline calls these methods before the plugin methods
    # are ever called
    def get_members(self, context, filters=None, fields=None):
        pass

    def get_member(self, context, id, fields=None):
        pass



# Copyright 2016 Radware LTD
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api import extensions as api_extensions
from neutron_lib import constants as n_constants
from neutron_lib.db import constants as db_const
from neutron_lib import exceptions as nexception
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory

from neutron.api import extensions
from neutron.api.v2 import base
from neutron.api.v2 import resource_helper

from neutron_lbaas._i18n import _
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.services.loadbalancer import constants as lb_const

LOADBALANCERV2_PREFIX = "/lbaas"


class L7PolicyRedirectPoolIdMissing(nexception.Conflict):
    message = _("Redirect pool id is missing for L7 Policy with"
                " pool redirect action.")


class L7PolicyRedirectUrlMissing(nexception.Conflict):
    message = _("Redirect URL is missing for L7 Policy with"
                " URL redirect action.")


class RuleNotFoundForL7Policy(nexception.NotFound):
    message = _("Rule %(rule_id)s could not be found in"
                " l7 policy %(l7policy_id)s.")


class L7RuleKeyMissing(nexception.NotFound):
    message = _("Rule key is missing."
                " Key should be specified for rules of "
                "HEADER and COOKIE types.")


class L7RuleInvalidKey(nexception.BadRequest):
    message = _("Invalid characters in key. See RFCs 2616, 2965, 6265, 7230.")


class L7RuleInvalidHeaderValue(nexception.BadRequest):
    message = _("Invalid characters in value. See RFC 7230.")


class L7RuleInvalidCookieValue(nexception.BadRequest):
    message = _("Invalid characters in value. See RFCs 2616, 2965, 6265.")


class L7RuleInvalidRegex(nexception.BadRequest):
    message = _("Unable to parse regular expression: %(e)s.")


class L7RuleUnsupportedCompareType(nexception.BadRequest):
    message = _("Unsupported compare type for rule of %(type)s type.")


RESOURCE_ATTRIBUTE_MAP = {
    'l7policies': {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True,
               'primary_key': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': db_const.NAME_FIELD_SIZE},
                      'required_by_policy': True,
                      'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'default': '',
                 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {
                            'type:string': db_const.DESCRIPTION_FIELD_SIZE},
                        'is_visible': True, 'default': ''},
        'listener_id': {'allow_post': True, 'allow_put': False,
                        'validate': {'type:uuid': None},
                        'is_visible': True},
        'action': {'allow_post': True, 'allow_put': True,
                   'validate': {
                       'type:values': lb_const.SUPPORTED_L7_POLICY_ACTIONS},
                   'is_visible': True},
        'redirect_pool_id': {'allow_post': True, 'allow_put': True,
                             'validate': {'type:uuid_or_none': None},
                             'default': n_constants.ATTR_NOT_SPECIFIED,
                             'is_visible': True},
        'redirect_url': {'allow_post': True, 'allow_put': True,
                         'validate': {
                             'type:regex_or_none': lb_const.URL_REGEX},
                         'default': None,
                         'is_visible': True},
        # range max is (2^31 - 1) to get around MySQL quirk
        'position': {'allow_post': True, 'allow_put': True,
                     'convert_to': converters.convert_to_int,
                     'validate': {'type:range': [1, 2147483647]},
                     'default': 2147483647,
                     'is_visible': True},
        'rules': {'allow_post': False, 'allow_put': False,
                 'is_visible': True},
        'admin_state_up': {'allow_post': True, 'allow_put': True,
                           'default': True,
                           'convert_to': converters.convert_to_boolean,
                           'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True}
    }
}

SUB_RESOURCE_ATTRIBUTE_MAP = {
    'rules': {
        'parent': {'collection_name': 'l7policies',
                   'member_name': 'l7policy'},
        'parameters': {
            'id': {'allow_post': False, 'allow_put': False,
                   'validate': {'type:uuid': None},
                   'is_visible': True,
                   'primary_key': True},
            'tenant_id': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:string': None},
                          'required_by_policy': True,
                          'is_visible': True},
            'type': {'allow_post': True, 'allow_put': True,
                     'validate': {
                         'type:values': lb_const.SUPPORTED_L7_RULE_TYPES},
                     'is_visible': True},
            'compare_type': {'allow_post': True, 'allow_put': True,
                             'validate': {
                                 'type:values':
                                 lb_const.SUPPORTED_L7_RULE_COMPARE_TYPES},
                             'is_visible': True},
            'invert': {'allow_post': True, 'allow_put': True,
                       'default': False,
                       'convert_to': converters.convert_to_boolean,
                       'is_visible': True},
            'key': {'allow_post': True, 'allow_put': True,
                    'validate': {'type:string_or_none': None},
                    'default': None,
                    'is_visible': True},
            'value': {'allow_post': True, 'allow_put': True,
                      'validate': {'type:string': None},
                      'is_visible': True},
            'admin_state_up': {'allow_post': True, 'allow_put': True,
                               'default': True,
                               'convert_to': converters.convert_to_boolean,
                               'is_visible': True},
            'status': {'allow_post': False, 'allow_put': False,
                       'is_visible': True}
        }
    }
}


class L7(api_extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "L7 capabilities for LBaaSv2"

    @classmethod
    def get_alias(cls):
        return "l7"

    @classmethod
    def get_description(cls):
        return "Adding L7 policies and rules support for LBaaSv2"

    @classmethod
    def get_updated(cls):
        return "2016-01-24T10:00:00-00:00"

    def get_required_extensions(self):
        return ["lbaasv2"]

    @classmethod
    def get_resources(cls):
        l7_plurals = {'l7policies': 'l7policy', 'rules': 'rule'}

        plural_mappings = resource_helper.build_plural_mappings(
            l7_plurals, RESOURCE_ATTRIBUTE_MAP)

        resources = resource_helper.build_resource_info(
            plural_mappings,
            RESOURCE_ATTRIBUTE_MAP,
            constants.LOADBALANCERV2,
            register_quota=True)

        plugin = directory.get_plugin(constants.LOADBALANCERV2)

        for collection_name in SUB_RESOURCE_ATTRIBUTE_MAP:
            # Special handling needed for sub-resources with 'y' ending
            # (e.g. proxies -> proxy)
            resource_name = plural_mappings.get(collection_name,
                                                collection_name[:-1])
            parent = SUB_RESOURCE_ATTRIBUTE_MAP[collection_name].get('parent')
            params = SUB_RESOURCE_ATTRIBUTE_MAP[collection_name].get(
                'parameters')

            controller = base.create_resource(collection_name, resource_name,
                                              plugin, params,
                                              allow_bulk=True,
                                              parent=parent,
                                              allow_pagination=True,
                                              allow_sorting=True)

            resource = extensions.ResourceExtension(
                collection_name,
                controller, parent,
                path_prefix=LOADBALANCERV2_PREFIX,
                attr_map=params)
            resources.append(resource)

        return resources

    @classmethod
    def get_plugin_interface(cls):
        return loadbalancerv2.LoadBalancerPluginBaseV2

    def update_attributes_map(self, attributes, extension_attrs_map=None):
        super(L7, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}



# Copyright 2015 Radware LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lbaas._i18n import _
from neutron_lbaas.common import exceptions


class RadwareLBaasV2Exception(exceptions.LbaasException):
    message = _('An unknown exception occurred in '
               'Radware LBaaS v2 provider.')


class AuthenticationMissing(RadwareLBaasV2Exception):
    message = _('vDirect user/password missing. '
               'Specify in configuration file, under [radwarev2] section')


class WorkflowTemplateMissing(RadwareLBaasV2Exception):
    message = _('Workflow template %(workflow_template)s is missing '
               'on vDirect server. Upload missing workflow')


class RESTRequestFailure(RadwareLBaasV2Exception):
    message = _('REST request failed with status %(status)s. '
               'Reason: %(reason)s, Description: %(description)s. '
               'Success status codes are %(success_codes)s')


class UnsupportedEntityOperation(RadwareLBaasV2Exception):
    message = _('%(operation)s operation is not supported for %(entity)s.')



# Copyright 2015 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from neutron_lbaas.common import cert_manager
from neutron_lbaas.common.cert_manager import barbican_cert_manager as bcm
from neutron_lbaas.common.cert_manager import cert_manager as cmi
from neutron_lbaas.common.cert_manager import local_cert_manager as lcm
from neutron_lbaas.tests import base


class TestCertManager(base.BaseTestCase):

    def setUp(self):
        cert_manager._CERT_MANAGER_PLUGIN = None
        super(TestCertManager, self).setUp()

    def test_get_service_url(self):
        # Format: <servicename>://<region>/<resource>/<object_id>
        cfg.CONF.set_override('service_name',
                              'lbaas',
                              'service_auth')
        cfg.CONF.set_override('region',
                              'RegionOne',
                              'service_auth')
        self.assertEqual(
            'lbaas://RegionOne/loadbalancer/LB-ID',
            cmi.CertManager.get_service_url('LB-ID'))

    def test_barbican_cert_manager(self):
        cfg.CONF.set_override(
            'cert_manager_type',
            'barbican',
            group='certificates')
        self.assertEqual(cert_manager.get_backend().CertManager,
                         bcm.CertManager)

    def test_local_cert_manager(self):
        cfg.CONF.set_override(
            'cert_manager_type',
            'local',
            group='certificates')
        self.assertEqual(cert_manager.get_backend().CertManager,
                         lcm.CertManager)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import itertools

import neutron.conf.agent.common
import neutron.conf.services.provider_configuration

import neutron_lbaas.agent.agent
import neutron_lbaas.common.cert_manager
import neutron_lbaas.common.cert_manager.local_cert_manager
import neutron_lbaas.common.keystone
import neutron_lbaas.drivers.common.agent_driver_base
import neutron_lbaas.drivers.haproxy.namespace_driver
import neutron_lbaas.drivers.octavia.driver
import neutron_lbaas.drivers.radware.base_v2_driver
import neutron_lbaas.extensions.loadbalancerv2


def list_agent_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             neutron_lbaas.agent.agent.OPTS,
             neutron.conf.agent.common.INTERFACE_OPTS,
             neutron.conf.agent.common.INTERFACE_DRIVER_OPTS)
         ),
        ('haproxy',
         neutron_lbaas.drivers.haproxy.namespace_driver.OPTS)
    ]


def list_opts():
    return [
        ('DEFAULT',
         neutron_lbaas.drivers.common.agent_driver_base.AGENT_SCHEDULER_OPTS),
        ('quotas',
         neutron_lbaas.extensions.loadbalancerv2.lbaasv2_quota_opts),
        ('service_auth',
         neutron_lbaas.common.keystone.OPTS),
        ('service_providers',
         neutron.conf.services.provider_configuration.serviceprovider_opts),
        ('certificates',
         itertools.chain(
             neutron_lbaas.common.cert_manager.cert_manager_opts,
             neutron_lbaas.common.cert_manager.local_cert_manager.
             local_cert_manager_opts)
         )
    ]


def list_service_opts():
    return [
        ('radwarev2',
         neutron_lbaas.drivers.radware.base_v2_driver.driver_opts),
        ('radwarev2_debug',
         neutron_lbaas.drivers.radware.base_v2_driver.driver_debug_opts),
        ('haproxy',
         itertools.chain(
             neutron.conf.agent.common.INTERFACE_DRIVER_OPTS,
             neutron_lbaas.agent.agent.OPTS,
         )),
        ('octavia',
         neutron_lbaas.drivers.octavia.driver.OPTS)
    ]



# Copyright 2016 Blue Box, an IBM Company
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import extensions
from neutron_lib import exceptions as nexception

from neutron_lbaas._i18n import _


class ListenerPoolLoadbalancerMismatch(nexception.BadRequest):
    message = _("Pool %(pool_id)s is on loadbalancer %(lb_id)s.")


class ListenerDefaultPoolAlreadySet(nexception.InUse):
    message = _("Listener %(listener_id)s "
                "is already using default pool %(pool_id)s.")


class PoolMustHaveLoadbalancer(nexception.BadRequest):
    message = _("Pool must be created with a loadbalancer or listener.")


class ListenerMustHaveLoadbalancer(nexception.BadRequest):
    message = _("Listener must be created with a loadbalancer or pool.")


class ListenerAndPoolMustBeOnSameLoadbalancer(nexception.BadRequest):
    message = _("Listener and pool must be on the same loadbalancer.")


EXTENDED_ATTRIBUTES_2_0 = {
    'loadbalancers': {
        'pools': {'allow_post': False, 'allow_put': False,
                  'is_visible': True}},
    'listeners': {
        'loadbalancer_id': {'allow_post': True, 'allow_put': False,
                            'validate': {'type:uuid_or_none': None},
                            'default': None,
                            'is_visible': False},
        'default_pool_id': {'allow_post': True, 'allow_put': True,
                            'default': None,
                            'validate': {'type:uuid_or_none': None},
                            'is_visible': True}},
    'pools': {
        'listener_id': {'allow_post': True, 'allow_put': False,
                        'validate': {'type:uuid_or_none': None},
                        'default': None,
                        'is_visible': False},
        'loadbalancer_id': {'allow_post': True, 'allow_put': False,
                            'validate': {'type:uuid_or_none': None},
                            'default': None,
                            'is_visible': True},
        'loadbalancers': {'allow_post': False, 'allow_put': False,
                         'is_visible': True}}}


class Sharedpools(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Shared pools for LBaaSv2"

    @classmethod
    def get_alias(cls):
        return "shared_pools"

    @classmethod
    def get_description(cls):
        return "Allow pools to be shared among listeners for LBaaSv2"

    @classmethod
    def get_updated(cls):
        return "2016-01-20T10:00:00-00:00"

    def get_required_extensions(self):
        return ["lbaasv2"]

    def get_extended_resources(self, version):
        if version == "2.0":
            return dict(EXTENDED_ATTRIBUTES_2_0.items())
        else:
            return {}



# Copyright (c) 2013 OpenStack Foundation.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import faults
from neutron_lib import exceptions as nexception
from neutron_lib.exceptions import agent as agent_exc
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory

from neutron.api import extensions
from neutron.api.v2 import resource
from neutron import policy
from neutron import wsgi

from neutron_lbaas._i18n import _
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.services.loadbalancer import constants as lb_const

LOADBALANCER = 'agent-loadbalancer'
LOADBALANCERS = LOADBALANCER + 's'
LOADBALANCER_AGENT = 'loadbalancer-hosting-agent'


class LoadBalancerSchedulerController(wsgi.Controller):
    def index(self, request, **kwargs):
        lbaas_plugin = directory.get_plugin(plugin_const.LOADBALANCERV2)
        if not lbaas_plugin:
            return {'load_balancers': []}

        policy.enforce(request.context,
                       "get_%s" % LOADBALANCERS,
                       {},
                       plugin=lbaas_plugin)
        lbs = lbaas_plugin.db.list_loadbalancers_on_lbaas_agent(
            request.context, kwargs['agent_id'])
        return {'loadbalancers': [lb.to_api_dict() for lb in lbs]}


class LbaasAgentHostingLoadBalancerController(wsgi.Controller):
    def index(self, request, **kwargs):
        lbaas_plugin = directory.get_plugin(plugin_const.LOADBALANCERV2)
        if not lbaas_plugin:
            return

        policy.enforce(request.context,
                       "get_%s" % LOADBALANCER_AGENT,
                       {},
                       plugin=lbaas_plugin)
        return lbaas_plugin.db.get_agent_hosting_loadbalancer(
            request.context, kwargs['loadbalancer_id'])


class Lbaas_agentschedulerv2(api_extensions.ExtensionDescriptor):
    """Extension class supporting LBaaS agent scheduler.
    """

    @classmethod
    def get_name(cls):
        return "Loadbalancer Agent Scheduler V2"

    @classmethod
    def get_alias(cls):
        return lb_const.LBAAS_AGENT_SCHEDULER_V2_EXT_ALIAS

    @classmethod
    def get_description(cls):
        return "Schedule load balancers among lbaas agents"

    @classmethod
    def get_updated(cls):
        return "2013-02-07T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        parent = dict(member_name="agent",
                      collection_name="agents")

        controller = resource.Resource(LoadBalancerSchedulerController(),
                                       faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(
            LOADBALANCERS, controller, parent))

        parent = dict(member_name="loadbalancer",
                      collection_name="loadbalancers")

        controller = resource.Resource(
            LbaasAgentHostingLoadBalancerController(), faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(
            LOADBALANCER_AGENT, controller, parent,
            path_prefix=loadbalancerv2.LOADBALANCERV2_PREFIX))
        return exts

    def get_extended_resources(self, version):
        return {}


class NoEligibleBackend(nexception.NotFound):
    message = _("No eligible backend for pool %(pool_id)s")


class NoEligibleLbaasAgent(NoEligibleBackend):
    message = _("No eligible agent found "
                "for loadbalancer %(loadbalancer_id)s.")


class NoActiveLbaasAgent(agent_exc.AgentNotFound):
    message = _("No active agent found "
                "for loadbalancer %(loadbalancer_id)s.")


class LbaasAgentSchedulerPluginBase(object):
    """REST API to operate the lbaas agent scheduler.

    All of method must be in an admin context.
    """

    @abc.abstractmethod
    def list_loadbalancers_on_lbaas_agent(self, context, id):
        pass

    @abc.abstractmethod
    def get_agent_hosting_loadbalancer(self, context, loadbalancer_id,
                                       active=None):
        pass



# Copyright 2015, Radware LTD. All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ssl

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_serialization import base64
from oslo_serialization import jsonutils
from six.moves import http_client

from neutron_lbaas.drivers.radware import exceptions as r_exc

LOG = logging.getLogger(__name__)

RESP_STATUS = 0
RESP_REASON = 1
RESP_STR = 2
RESP_DATA = 3


class vDirectRESTClient(object):
    """REST server proxy to Radware vDirect."""
    @log_helpers.log_method_call
    def __init__(self,
                 server='localhost',
                 secondary_server=None,
                 user=None,
                 password=None,
                 port=2189,
                 ssl=True,
                 ssl_verify_context=True,
                 timeout=5000,
                 base_uri=''):
        self.server = server
        self.secondary_server = secondary_server
        self.port = port
        self.ssl = ssl
        self.ssl_verify_context = ssl_verify_context
        self.base_uri = base_uri
        self.timeout = timeout
        if user and password:
            self.auth = base64.encode_as_text('%s:%s' % (user, password))
            self.auth = self.auth.replace('\n', '')
        else:
            raise r_exc.AuthenticationMissing()

        debug_params = {'server': self.server,
                        'sec_server': self.secondary_server,
                        'port': self.port,
                        'ssl': self.ssl,
                        'ssl_verify_context': self.ssl_verify_context}
        LOG.debug('vDirectRESTClient:init server=%(server)s, '
                  'secondary server=%(sec_server)s, '
                  'port=%(port)d, '
                  'ssl=%(ssl)r, '
                  'ssl_verify_context=%(ssl_verify_context)r' %
                  debug_params)

    def _flip_servers(self):
        LOG.warning('Flipping servers. Current is: %(server)s, switching to '
                    '%(secondary)s', {'server': self.server,
                                      'secondary': self.secondary_server})
        self.server, self.secondary_server = self.secondary_server, self.server

    def _recover(self, action, resource, data, headers, binary=False):
        if self.server and self.secondary_server:
            self._flip_servers()
            resp = self._call(action, resource, data,
                              headers, binary)
            return resp
        else:
            LOG.error('REST client is not able to recover '
                      'since only one vDirect server is '
                      'configured.')
            return -1, None, None, None

    def call(self, action, resource, data, headers, binary=False):
        resp = self._call(action, resource, data, headers, binary)
        if resp[RESP_STATUS] == -1:
            LOG.warning('vDirect server is not responding (%s).',
                        self.server)
            return self._recover(action, resource, data, headers, binary)
        elif resp[RESP_STATUS] in (301, 307):
            LOG.warning('vDirect server is not active (%s).',
                        self.server)
            return self._recover(action, resource, data, headers, binary)
        else:
            return resp

    @log_helpers.log_method_call
    def _call(self, action, resource, data, headers, binary=False):
        if resource.startswith('http'):
            uri = resource
        else:
            uri = self.base_uri + resource
        if binary:
            body = data
        else:
            body = jsonutils.dumps(data)

        debug_data = 'binary' if binary else body
        debug_data = debug_data if debug_data else 'EMPTY'
        if not headers:
            headers = {'Authorization': 'Basic %s' % self.auth}
        else:
            headers['Authorization'] = 'Basic %s' % self.auth
        conn = None
        if self.ssl:
            if not self.ssl_verify_context:
                # HTTPS certificate verification was cancelled in
                # configuration. If python version has context attribute,
                # switch the default context to unverified
                try:
                    _create_unverified_https_context =\
                        ssl._create_unverified_context
                except AttributeError:
                    # Legacy Python that doesn't verify HTTPS
                    # certificates by default
                    pass
                else:
                    # Handle target environment that doesn't
                    # support HTTPS verification
                    ssl._create_default_https_context =\
                        _create_unverified_https_context

            conn = http_client.HTTPSConnection(
                self.server, self.port, timeout=self.timeout)
            if conn is None:
                LOG.error('vdirectRESTClient: Could not establish HTTPS '
                          'connection')
                return 0, None, None, None
        else:
            conn = http_client.HTTPConnection(
                self.server, self.port, timeout=self.timeout)
            if conn is None:
                LOG.error('vdirectRESTClient: Could not establish HTTP '
                          'connection')
                return 0, None, None, None

        try:
            conn.request(action, uri, body, headers)
            response = conn.getresponse()
            respstr = response.read()
            respdata = respstr
            try:
                respdata = jsonutils.loads(respstr)
            except ValueError:
                # response was not JSON, ignore the exception
                pass
            ret = (response.status, response.reason, respstr, respdata)
        except Exception as e:
            log_dict = {'action': action, 'e': e}
            LOG.error('vdirectRESTClient: %(action)s failure, %(e)r',
                      log_dict)
            ret = -1, None, None, None
        conn.close()
        return ret




# Copyright 2014 Citrix Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_serialization import jsonutils

from neutron_lbaas._i18n import _

LOG = logging.getLogger(__name__)

CONTENT_TYPE_HEADER = 'Content-type'
ACCEPT_HEADER = 'Accept'
AUTH_HEADER = 'Cookie'
DRIVER_HEADER = 'X-OpenStack-LBaaS'
TENANT_HEADER = 'X-Tenant-ID'
JSON_CONTENT_TYPE = 'application/json'
DRIVER_HEADER_VALUE = 'netscaler-openstack-lbaas'
NITRO_LOGIN_URI = 'nitro/v2/config/login'


class NCCException(n_exc.NeutronException):

    """Represents exceptions thrown by NSClient."""

    CONNECTION_ERROR = 1
    REQUEST_ERROR = 2
    RESPONSE_ERROR = 3
    UNKNOWN_ERROR = 4

    def __init__(self, error, status=requests.codes.SERVICE_UNAVAILABLE):
        self.message = _("NCC Error %d") % error
        super(NCCException, self).__init__()
        self.error = error
        self.status = status

    def is_not_found_exception(self):
        if int(self.status) == requests.codes.NOT_FOUND:
            return True


class NSClient(object):

    """Client to operate on REST resources of NetScaler Control Center."""

    def __init__(self, service_uri, username, password,
                 ncc_cleanup_mode="False"):
        if not service_uri:
            LOG.exception("No NetScaler Control Center URI specified. "
                          "Cannot connect.")
            raise NCCException(NCCException.CONNECTION_ERROR)
        self.service_uri = service_uri.strip('/')
        self.auth = None
        self.cleanup_mode = False
        if username and password:
            self.username = username
            self.password = password
        if ncc_cleanup_mode.lower() == "true":
            self.cleanup_mode = True

    def create_resource(self, tenant_id, resource_path, object_name,
                        object_data):
        """Create a resource of NetScaler Control Center."""
        return self._resource_operation('POST', tenant_id,
                                        resource_path,
                                        object_name=object_name,
                                        object_data=object_data)

    def is_login(self, resource_uri):
        if 'login' in resource_uri.lower():
            return True
        else:
            return False

    def login(self):
        """Get session based login"""
        login_obj = {"username": self.username, "password": self.password}

        msg = "NetScaler driver login:" + repr(self.username)
        LOG.info(msg)
        resp_status, result = self.create_resource("login", NITRO_LOGIN_URI,
                                                   "login", login_obj)
        LOG.info("Response: status : %(status)s %result(result)s", {
                "status": resp_status, "result": result['body']})
        result_body = jsonutils.loads(result['body'])

        session_id = None
        if result_body and "login" in result_body:
            logins = result_body["login"]
            if isinstance(logins, list):
                login = logins[0]
            else:
                login = logins
            if login and "sessionid" in login:
                session_id = login["sessionid"]

        if session_id:
            LOG.info("Response: %(result)s", {"result": result['body']})
            LOG.info(
                "Session_id = %(session_id)s" %
                {"session_id": session_id})
            # Update sessin_id in auth
            self.auth = "SessId=%s" % session_id
        else:
            raise NCCException(NCCException.RESPONSE_ERROR)

    def retrieve_resource(self, tenant_id, resource_path, parse_response=True):
        """Retrieve a resource of NetScaler Control Center."""
        return self._resource_operation('GET', tenant_id, resource_path)

    def update_resource(self, tenant_id, resource_path, object_name,
                        object_data):
        """Update a resource of the NetScaler Control Center."""
        return self._resource_operation('PUT', tenant_id,
                                        resource_path,
                                        object_name=object_name,
                                        object_data=object_data)

    def remove_resource(self, tenant_id, resource_path, parse_response=True):
        """Remove a resource of NetScaler Control Center."""
        if self.cleanup_mode:
            return True
        else:
            return self._resource_operation('DELETE', tenant_id, resource_path)

    def _resource_operation(self, method, tenant_id, resource_path,
                            object_name=None, object_data=None):
        resource_uri = "%s/%s" % (self.service_uri, resource_path)
        if not self.auth and not self.is_login(resource_uri):
            # Creating a session for the first time
            self.login()
        headers = self._setup_req_headers(tenant_id)
        request_body = None
        if object_data:
            if isinstance(object_data, str):
                request_body = object_data
            else:
                obj_dict = {object_name: object_data}
                request_body = jsonutils.dumps(obj_dict)
        try:
            response_status, resp_dict = (self.
                                          _execute_request(method,
                                                           resource_uri,
                                                           headers,
                                                           body=request_body))
        except NCCException as e:
            if e.status == requests.codes.NOT_FOUND and method == 'DELETE':
                return 200, {}
            else:
                raise e

        return response_status, resp_dict

    def _is_valid_response(self, response_status):
        # when status is less than 400, the response is fine
        return response_status < requests.codes.bad_request

    def _setup_req_headers(self, tenant_id):
        headers = {ACCEPT_HEADER: JSON_CONTENT_TYPE,
                   CONTENT_TYPE_HEADER: JSON_CONTENT_TYPE,
                   DRIVER_HEADER: DRIVER_HEADER_VALUE,
                   TENANT_HEADER: tenant_id,
                   AUTH_HEADER: self.auth}
        return headers

    def _get_response_dict(self, response):
        response_dict = {'status': int(response.status_code),
                         'body': response.text,
                         'headers': response.headers}
        if self._is_valid_response(int(response.status_code)):
            if response.text:
                response_dict['dict'] = response.json()
        return response_dict

    def _execute_request(self, method, resource_uri, headers, body=None):
        service_uri_dict = {"service_uri": self.service_uri}
        try:
            response = requests.request(method, url=resource_uri,
                                        headers=headers, data=body)
        except requests.exceptions.SSLError:
            LOG.exception("SSL error occurred while connecting "
                         "to %(service_uri)s",
                         service_uri_dict)
            raise NCCException(NCCException.CONNECTION_ERROR)
        except requests.exceptions.ConnectionError:
            LOG.exception("Connection error occurred while connecting"
                         "to %(service_uri)s", service_uri_dict)
            raise NCCException(NCCException.CONNECTION_ERROR)

        except requests.exceptions.Timeout:
            LOG.exception(
                "Request to %(service_uri)s timed out", service_uri_dict)
            raise NCCException(NCCException.CONNECTION_ERROR)
        except (requests.exceptions.URLRequired,
                requests.exceptions.InvalidURL,
                requests.exceptions.MissingSchema,
                requests.exceptions.InvalidSchema):
            LOG.exception("Request did not specify a valid URL")
            raise NCCException(NCCException.REQUEST_ERROR)
        except requests.exceptions.TooManyRedirects:
            LOG.exception("Too many redirects occurred for request ")
            raise NCCException(NCCException.REQUEST_ERROR)
        except requests.exceptions.RequestException:
            LOG.exception(
                "A request error while connecting to %(service_uri)s",
                service_uri_dict)
            raise NCCException(NCCException.REQUEST_ERROR)
        except Exception:
            LOG.exception(
                "A unknown error occurred during request to"
                " %(service_uri)s", service_uri_dict)
            raise NCCException(NCCException.UNKNOWN_ERROR)
        resp_dict = self._get_response_dict(response)
        resp_body = resp_dict['body']
        LOG.info("Response: %(resp_body)s", {"resp_body": resp_body})
        response_status = resp_dict['status']
        if response_status == requests.codes.unauthorized:
            LOG.exception("Unable to login. Invalid credentials passed."
                         "for: %s", self.service_uri)
            if not self.is_login(resource_uri):
                # Session expired, relogin and retry....
                self.login()
                # Retry the operation
                headers.update({AUTH_HEADER: self.auth})
                self._execute_request(method,
                                      resource_uri,
                                      headers,
                                      body)
            else:
                raise NCCException(NCCException.RESPONSE_ERROR)
        if not self._is_valid_response(response_status):
            response_msg = resp_body
            response_dict = {"method": method,
                             "url": resource_uri,
                             "response_status": response_status,
                             "response_msg": response_msg}
            LOG.exception("Failed %(method)s operation on %(url)s "
                         "status code: %(response_status)s "
                         "message: %(response_msg)s", response_dict)
            raise NCCException(NCCException.RESPONSE_ERROR, response_status)
        return response_status, resp_dict



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import uuidutils
from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base

CONF = config.CONF


class TestHealthMonitors(base.BaseAdminTestCase):

    """
    Tests the following operations in the Neutron-LBaaS API using the
    REST client for Health Monitors with ADMIN role:

    create health monitor with missing tenant_id
    create health monitor with empty tenant id
    create health monitor with another tenant_id
    """

    @classmethod
    def resource_setup(cls):
        super(TestHealthMonitors, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.load_balancer = cls._create_load_balancer(
            tenant_id=cls.subnet.get('tenant_id'),
            vip_subnet_id=cls.subnet.get('id'))
        cls.listener = cls._create_listener(
            loadbalancer_id=cls.load_balancer.get('id'),
            protocol=cls.listener_protocol, protocol_port=80)
        cls.pool = cls._create_pool(
            protocol=cls.pool_protocol, lb_algorithm='ROUND_ROBIN',
            listener_id=cls.listener.get('id'))

    @classmethod
    def resource_cleanup(cls):
        super(TestHealthMonitors, cls).resource_cleanup()

    def test_create_health_monitor_missing_tenant_id_field(self):
        """
        Test if admin user can create health monitor with a missing tenant id
        field.
        """
        hm = self._create_health_monitor(type=self.hm_protocol, delay=3,
                                         max_retries=10, timeout=5,
                                         pool_id=self.pool.get('id'))

        admin_hm = self.health_monitors_client.get_health_monitor(hm.get('id'))
        admin_tenant_id = admin_hm.get('tenant_id')
        hm_tenant_id = hm.get('tenant_id')
        self.assertEqual(admin_tenant_id, hm_tenant_id)

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_health_monitor_empty_tenant_id_field(self):
        """
        Test with admin user creating health monitor with an empty tenant id
        field should fail.
        """
        self.assertRaises(ex.BadRequest, self._create_health_monitor,
                          type=self.hm_protocol, delay=3, max_retries=10,
                          timeout=5, pool_id=self.pool.get('id'), tenant_id="")

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='smoke')
    def test_create_health_monitor_for_another_tenant_id_field(self):
        """Test with admin user create health monitor for another tenant id.
        """

        tenantid = uuidutils.generate_uuid()
        hm = self._create_health_monitor(type=self.hm_protocol, delay=3,
                                         max_retries=10, timeout=5,
                                         pool_id=self.pool.get('id'),
                                         tenant_id=tenantid)

        self.assertEqual(hm.get('tenant_id'), tenantid)
        self.assertNotEqual(hm.get('tenant_id'),
                            self.subnet.get('tenant_id'))



# Copyright 2013 New Dream Network, LLC (DreamHost)
# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import rpc as n_rpc
import oslo_messaging


class LbaasAgentApi(object):
    """Agent side of the Agent to Plugin RPC API."""

    # history
    #   1.0 Initial version

    def __init__(self, topic, context, host):
        self.context = context
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def get_ready_devices(self):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'get_ready_devices', host=self.host)

    def get_loadbalancer(self, loadbalancer_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'get_loadbalancer',
                          loadbalancer_id=loadbalancer_id)

    def loadbalancer_deployed(self, loadbalancer_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'loadbalancer_deployed',
                          loadbalancer_id=loadbalancer_id)

    def update_status(self, obj_type, obj_id, provisioning_status=None,
                      operating_status=None):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'update_status', obj_type=obj_type,
                          obj_id=obj_id,
                          provisioning_status=provisioning_status,
                          operating_status=operating_status)

    def loadbalancer_destroyed(self, loadbalancer_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'loadbalancer_destroyed',
                          loadbalancer_id=loadbalancer_id)

    def plug_vip_port(self, port_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'plug_vip_port', port_id=port_id,
                          host=self.host)

    def unplug_vip_port(self, port_id):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'unplug_vip_port', port_id=port_id,
                          host=self.host)

    def update_loadbalancer_stats(self, loadbalancer_id, stats):
        cctxt = self.client.prepare()
        return cctxt.call(self.context, 'update_loadbalancer_stats',
                          loadbalancer_id=loadbalancer_id, stats=stats)



# Copyright 2013 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins import constants as n_constants
from neutron_lib.plugins import directory
from oslo_utils import importutils

from neutron.db import servicetype_db as st_db
from neutron.tests.common import helpers

from neutron_lbaas.common import exceptions
from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.drivers.common import agent_driver_base
from neutron_lbaas.extensions import loadbalancerv2
from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer.data_models import LoadBalancer
from neutron_lbaas.tests import base
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2
from neutron_lbaas.tests.unit import test_agent_scheduler


class TestLoadBalancerPluginBase(test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def setUp(self):
        def reset_device_driver():
            agent_driver_base.AgentDriverBase.device_driver = None
        self.addCleanup(reset_device_driver)

        self.mock_importer = mock.patch.object(
            agent_driver_base, 'importutils').start()

        # needed to reload provider configuration
        st_db.ServiceTypeManager._instance = None
        agent_driver_base.AgentDriverBase.device_driver = 'dummy'
        super(TestLoadBalancerPluginBase, self).setUp(
            lbaas_provider=('LOADBALANCERV2:lbaas:neutron_lbaas.drivers.'
                            'common.agent_driver_base.'
                            'AgentDriverBase:default'))

        # we need access to loaded plugins to modify models
        self.plugin_instance = directory.get_plugin(n_constants.LOADBALANCERV2)


class TestLoadBalancerAgentApi(base.BaseTestCase):
    def setUp(self):
        super(TestLoadBalancerAgentApi, self).setUp()

        self.api = agent_driver_base.LoadBalancerAgentApi('topic')

    def test_init(self):
        self.assertEqual('topic', self.api.client.target.topic)

    def _call_test_helper(self, method_name, method_args):
        with mock.patch.object(self.api.client, 'cast') as rpc_mock, \
                mock.patch.object(self.api.client, 'prepare') as prepare_mock:
            prepare_mock.return_value = self.api.client
            getattr(self.api, method_name)(mock.sentinel.context,
                                           host='host',
                                           **method_args)

        prepare_args = {'server': 'host'}
        prepare_mock.assert_called_once_with(**prepare_args)

        if method_name == 'agent_updated':
            method_args = {'payload': method_args}
        rpc_mock.assert_called_once_with(mock.sentinel.context, method_name,
                                         **method_args)

    def test_agent_updated(self):
        self._call_test_helper('agent_updated', {'admin_state_up': 'test'})

    def test_create_pool(self):
        self._call_test_helper('create_pool', {'pool': 'test'})

    def test_update_pool(self):
        self._call_test_helper('update_pool', {'old_pool': 'test',
                                               'pool': 'test'})

    def test_delete_pool(self):
        self._call_test_helper('delete_pool', {'pool': 'test'})

    def test_create_loadbalancer(self):
        self._call_test_helper('create_loadbalancer', {'loadbalancer': 'test',
                                                       'driver_name': 'dummy'})

    def test_update_loadbalancer(self):
        self._call_test_helper('update_loadbalancer', {
            'old_loadbalancer': 'test', 'loadbalancer': 'test'})

    def test_delete_loadbalancer(self):
        self._call_test_helper('delete_loadbalancer', {'loadbalancer': 'test'})

    def test_create_member(self):
        self._call_test_helper('create_member', {'member': 'test'})

    def test_update_member(self):
        self._call_test_helper('update_member', {'old_member': 'test',
                                                 'member': 'test'})

    def test_delete_member(self):
        self._call_test_helper('delete_member', {'member': 'test'})

    def test_create_monitor(self):
        self._call_test_helper('create_healthmonitor',
                               {'healthmonitor': 'test'})

    def test_update_monitor(self):
        self._call_test_helper('update_healthmonitor',
                               {'old_healthmonitor': 'test',
                                'healthmonitor': 'test'})

    def test_delete_monitor(self):
        self._call_test_helper('delete_healthmonitor',
                               {'healthmonitor': 'test'})


class TestLoadBalancerPluginNotificationWrapper(TestLoadBalancerPluginBase):
    def setUp(self):
        self.log = mock.patch.object(agent_driver_base, 'LOG')
        api_cls = mock.patch.object(agent_driver_base,
                                    'LoadBalancerAgentApi').start()
        super(TestLoadBalancerPluginNotificationWrapper, self).setUp()
        self.mock_api = api_cls.return_value

        self.mock_get_driver = mock.patch.object(self.plugin_instance,
                                                 '_get_driver')
        self.mock_get_driver.return_value = (
            agent_driver_base.AgentDriverBase(self.plugin_instance))

    def _update_status(self, model, status, id):
        ctx = context.get_admin_context()
        self.plugin_instance.db.update_status(
            ctx,
            model,
            id,
            provisioning_status=status
        )

    def test_create_loadbalancer(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            calls = self.mock_api.create_loadbalancer.call_args_list
            self.assertEqual(1, len(calls))
            _, called_lb, _, device_driver = calls[0][0]
            self.assertEqual(loadbalancer['loadbalancer']['id'], called_lb.id)
            self.assertEqual('dummy', device_driver)
            self.assertEqual(constants.PENDING_CREATE,
                             called_lb.provisioning_status)

    def test_update_loadbalancer(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            old_lb_name = loadbalancer['loadbalancer']['name']
            ctx = context.get_admin_context()
            self.plugin_instance.db.update_loadbalancer_provisioning_status(
                ctx,
                loadbalancer['loadbalancer']['id'])
            new_lb_name = 'new_lb_name'
            loadbalancer['loadbalancer']['name'] = new_lb_name
            self._update_loadbalancer_api(
                lb_id, {'loadbalancer': {'name': new_lb_name}})
            calls = self.mock_api.update_loadbalancer.call_args_list
            self.assertEqual(1, len(calls))
            _, called_old_lb, called_new_lb, called_host = calls[0][0]
            self.assertEqual(lb_id, called_old_lb.id)
            self.assertEqual(lb_id, called_new_lb.id)
            self.assertEqual(old_lb_name, called_old_lb.name)
            self.assertEqual(new_lb_name, called_new_lb.name)
            self.assertEqual('host', called_host)
            self.assertEqual(constants.PENDING_UPDATE,
                             called_new_lb.provisioning_status)

    def test_delete_loadbalancer(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            ctx = context.get_admin_context()
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            self.plugin_instance.delete_loadbalancer(ctx, lb_id)
            calls = self.mock_api.delete_loadbalancer.call_args_list
            self.assertEqual(1, len(calls))
            _, called_lb, called_host = calls[0][0]
            self.assertEqual(lb_id, called_lb.id)
            self.assertEqual('host', called_host)
            self.assertEqual(constants.PENDING_DELETE,
                             called_lb.provisioning_status)
            self.assertRaises(loadbalancerv2.EntityNotFound,
                              self.plugin_instance.db.get_loadbalancer,
                              ctx, lb_id)

    def test_create_listener(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE,
                                loadbalancer['loadbalancer']['id'])
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                calls = self.mock_api.create_listener.call_args_list
                _, called_listener, called_host = calls[0][0]
                self.assertEqual(listener_id, called_listener.id)
                self.assertEqual('host', called_host)
                self.assertEqual(constants.PENDING_CREATE,
                                 called_listener.provisioning_status)
                ctx = context.get_admin_context()
                lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                self.assertEqual(constants.PENDING_UPDATE,
                                 lb.provisioning_status)

    def test_update_listener(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE,
                                loadbalancer['loadbalancer']['id'])
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                old_name = listener['listener']['name']
                ctx = context.get_admin_context()
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                self.plugin_instance.db.get_listener(ctx, listener_id)
                new_name = 'new_listener_name'
                listener['listener']['name'] = new_name
                self.plugin_instance.update_listener(
                    ctx, listener['listener']['id'], listener)
                self.plugin_instance.db.get_listener(
                    ctx, listener['listener']['id'])
                calls = self.mock_api.update_listener.call_args_list
                (_, old_called_listener,
                 new_called_listener, called_host) = calls[0][0]
                self.assertEqual(listener_id, new_called_listener.id)
                self.assertEqual(listener_id, old_called_listener.id)
                self.assertEqual(old_name, old_called_listener.name)
                self.assertEqual(new_name, new_called_listener.name)
                self.assertEqual(constants.PENDING_UPDATE,
                                 new_called_listener.provisioning_status)
                lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                self.assertEqual(constants.PENDING_UPDATE,
                                 lb.provisioning_status)
                self.assertEqual('host', called_host)

    def test_delete_listener(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                ctx = context.get_admin_context()
                self.plugin_instance.delete_listener(
                    ctx, listener['listener']['id'])
                calls = self.mock_api.delete_listener.call_args_list
                _, called_listener, called_host = calls[0][0]
                self.assertEqual(listener_id, called_listener.id)
                self.assertEqual('host', called_host)
                self.assertEqual(constants.PENDING_DELETE,
                                 called_listener.provisioning_status)
                ctx = context.get_admin_context()
                lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                self.assertEqual(constants.ACTIVE,
                                 lb.provisioning_status)
                self.assertRaises(
                    loadbalancerv2.EntityNotFound,
                    self.plugin_instance.db.get_listener, ctx, listener_id)

    def test_create_pool(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    calls = self.mock_api.create_pool.call_args_list
                    _, called_pool, called_host = calls[0][0]
                    self.assertEqual(pool_id, called_pool.id)
                    self.assertEqual('host', called_host)
                    self.assertEqual(constants.PENDING_CREATE,
                                     called_pool.provisioning_status)
                    ctx = context.get_admin_context()
                    lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                    self.assertEqual(constants.PENDING_UPDATE,
                                     lb.provisioning_status)

    def test_update_pool(self):
        ctx = context.get_admin_context()
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(loadbalancer_id=lb_id, listener_id=listener_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    old_name = pool['pool']['name']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    new_name = 'new_name'
                    pool['pool']['name'] = new_name
                    self.plugin_instance.update_pool(ctx, pool_id, pool)
                    calls = self.mock_api.update_pool.call_args_list
                    (_, old_called_pool,
                     new_called_pool, called_host) = calls[0][0]
                    self.assertEqual(pool_id, new_called_pool.id)
                    self.assertEqual(pool_id, old_called_pool.id)
                    self.assertEqual(old_name, old_called_pool.name)
                    self.assertEqual(new_name, new_called_pool.name)
                    self.assertEqual(constants.PENDING_UPDATE,
                                     new_called_pool.provisioning_status)
                    lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                    self.assertEqual(constants.PENDING_UPDATE,
                                     lb.provisioning_status)
                    self.assertEqual('host', called_host)

    def test_delete_pool(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    ctx = context.get_admin_context()
                    self.plugin_instance.delete_pool(ctx, pool_id)
                    calls = self.mock_api.delete_pool.call_args_list
                    _, called_pool, called_host = calls[0][0]
                    self.assertEqual(pool_id, called_pool.id)
                    self.assertEqual('host', called_host)
                    self.assertEqual(constants.PENDING_DELETE,
                                     called_pool.provisioning_status)
                    lb = self.plugin_instance.db.get_loadbalancer(ctx, lb_id)
                    self.assertEqual(constants.ACTIVE,
                                     lb.provisioning_status)
                    self.assertRaises(
                        loadbalancerv2.EntityNotFound,
                        self.plugin_instance.db.get_pool, ctx, pool_id)

    def test_create_member(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.subnet(cidr='11.0.0.0/24') as subnet:
                        with self.member(pool_id=pool_id, subnet=subnet,
                                         no_delete=True) as member:
                            member_id = member['member']['id']
                            calls = self.mock_api.create_member.call_args_list
                            _, called_member, called_host = calls[0][0]
                            self.assertEqual(member_id, called_member.id)
                            self.assertEqual('host', called_host)
                            self.assertEqual(constants.PENDING_CREATE,
                                             called_member.provisioning_status)
                            ctx = context.get_admin_context()
                            lb = self.plugin_instance.db.get_loadbalancer(
                                ctx, lb_id)
                            self.assertEqual(constants.PENDING_UPDATE,
                                             lb.provisioning_status)

    def test_update_member(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.subnet(cidr='11.0.0.0/24') as subnet:
                        with self.member(pool_id=pool_id, subnet=subnet,
                                         no_delete=True) as member:
                            member_id = member['member']['id']
                            self._update_status(models.LoadBalancer,
                                                constants.ACTIVE, lb_id)
                            old_weight = member['member']['weight']
                            new_weight = 2
                            member['member']['weight'] = new_weight
                            ctx = context.get_admin_context()
                            self.plugin_instance.update_pool_member(
                                ctx, member_id, pool_id, member)
                            calls = self.mock_api.update_member.call_args_list
                            (_, old_called_member,
                             new_called_member, called_host) = calls[0][0]
                            self.assertEqual(member_id, new_called_member.id)
                            self.assertEqual(member_id, old_called_member.id)
                            self.assertEqual(old_weight,
                                             old_called_member.weight)
                            self.assertEqual(new_weight,
                                             new_called_member.weight)
                            self.assertEqual(
                                constants.PENDING_UPDATE,
                                new_called_member.provisioning_status)
                            lb = self.plugin_instance.db.get_loadbalancer(
                                ctx, lb_id)
                            self.assertEqual(constants.PENDING_UPDATE,
                                             lb.provisioning_status)
                            self.assertEqual('host', called_host)

    def test_delete_member(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.subnet(cidr='11.0.0.0/24') as subnet:
                        with self.member(pool_id=pool_id, subnet=subnet,
                                         no_delete=True) as member:
                            member_id = member['member']['id']
                            self._update_status(models.LoadBalancer,
                                                constants.ACTIVE, lb_id)
                            ctx = context.get_admin_context()
                            self.plugin_instance.delete_pool_member(
                                ctx, member_id, pool_id)
                            calls = self.mock_api.delete_member.call_args_list
                            _, called_member, called_host = calls[0][0]
                            self.assertEqual(member_id, called_member.id)
                            self.assertEqual('host', called_host)
                            self.assertEqual(constants.PENDING_DELETE,
                                             called_member.provisioning_status)
                            lb = self.plugin_instance.db.get_loadbalancer(
                                ctx, lb_id)
                            self.assertEqual(constants.ACTIVE,
                                             lb.provisioning_status)
                            self.assertRaises(
                                loadbalancerv2.EntityNotFound,
                                self.plugin_instance.db.get_pool_member,
                                ctx, member_id)

    def test_create_health_monitor(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.healthmonitor(pool_id=pool_id,
                                            no_delete=True) as monitor:
                        hm_id = monitor['healthmonitor']['id']
                        calls = (
                            self.mock_api.create_healthmonitor.call_args_list)
                        _, called_hm, called_host = calls[0][0]
                        self.assertEqual(hm_id, called_hm.id)
                        self.assertEqual('host', called_host)
                        self.assertEqual(constants.PENDING_CREATE,
                                         called_hm.provisioning_status)
                        ctx = context.get_admin_context()
                        lb = self.plugin_instance.db.get_loadbalancer(
                            ctx, lb_id)
                        self.assertEqual(constants.PENDING_UPDATE,
                                         lb.provisioning_status)

    def test_update_health_monitor(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.healthmonitor(pool_id=pool_id,
                                            no_delete=True) as monitor:
                        hm_id = monitor['healthmonitor']['id']
                        self._update_status(models.LoadBalancer,
                                            constants.ACTIVE, lb_id)
                        old_to = monitor['healthmonitor']['timeout']
                        new_to = 2
                        monitor['healthmonitor']['timeout'] = new_to
                        ctx = context.get_admin_context()
                        self.plugin_instance.update_healthmonitor(ctx, hm_id,
                                                                  monitor)
                        calls = (
                            self.mock_api.update_healthmonitor.call_args_list)
                        (_, old_called_hm,
                         new_called_hm, called_host) = calls[0][0]
                        self.assertEqual(hm_id, new_called_hm.id)
                        self.assertEqual(hm_id, old_called_hm.id)
                        self.assertEqual(old_to,
                                         old_called_hm.timeout)
                        self.assertEqual(new_to,
                                         new_called_hm.timeout)
                        self.assertEqual(
                            constants.PENDING_UPDATE,
                            new_called_hm.provisioning_status)
                        lb = self.plugin_instance.db.get_loadbalancer(
                            ctx, lb_id)
                        self.assertEqual(constants.PENDING_UPDATE,
                                         lb.provisioning_status)
                        self.assertEqual('host', called_host)

    def test_delete_health_monitor(self):
        with self.loadbalancer(no_delete=True) as loadbalancer:
            lb_id = loadbalancer['loadbalancer']['id']
            self._update_status(models.LoadBalancer, constants.ACTIVE, lb_id)
            with self.listener(loadbalancer_id=lb_id,
                               no_delete=True) as listener:
                listener_id = listener['listener']['id']
                self._update_status(models.LoadBalancer, constants.ACTIVE,
                                    lb_id)
                with self.pool(listener_id=listener_id, loadbalancer_id=lb_id,
                               no_delete=True) as pool:
                    pool_id = pool['pool']['id']
                    self._update_status(models.LoadBalancer, constants.ACTIVE,
                                        lb_id)
                    with self.healthmonitor(pool_id=pool_id,
                                            no_delete=True) as monitor:
                        hm_id = monitor['healthmonitor']['id']
                        self._update_status(models.LoadBalancer,
                                            constants.ACTIVE, lb_id)
                        ctx = context.get_admin_context()
                        self.plugin_instance.delete_healthmonitor(ctx, hm_id)
                        calls = (
                            self.mock_api.delete_healthmonitor.call_args_list)
                        _, called_hm, called_host = calls[0][0]
                        self.assertEqual(hm_id, called_hm.id)
                        self.assertEqual('host', called_host)
                        self.assertEqual(constants.PENDING_DELETE,
                                         called_hm.provisioning_status)
                        lb = self.plugin_instance.db.get_loadbalancer(
                            ctx, lb_id)
                        self.assertEqual(constants.ACTIVE,
                                         lb.provisioning_status)
                        self.assertRaises(
                            loadbalancerv2.EntityNotFound,
                            self.plugin_instance.db.get_healthmonitor,
                            ctx, hm_id)


class TestLoadBalancerManager(test_agent_scheduler.
                              LBaaSAgentSchedulerTestCase):

    def setUp(self):
        super(TestLoadBalancerManager, self).setUp()
        self.load_balancer = agent_driver_base.LoadBalancerManager(self)
        self.agent_rpc = agent_driver_base.LoadBalancerAgentApi(
                lb_const.LOADBALANCER_AGENTV2)
        self.plugin = self.lbaas_plugin
        self.device_driver = 'haproxy_ns'
        self.loadbalancer_scheduler = importutils.import_object(
            'neutron_lbaas.agent_scheduler.ChanceScheduler')

    def test_reschedule_lbaas_from_down_agents(self):
        with mock.patch(
            'neutron_lbaas.drivers.common.agent_driver_base.'
            'LoadBalancerManager.reschedule_resources_from_down_agents'
        ) as mock_reschedule_resources:
            self.load_balancer.reschedule_lbaas_from_down_agents()
            self.assertTrue(mock_reschedule_resources.called)
            mock_reschedule_resources.assert_called_once_with(
                agent_type=lb_const.AGENT_TYPE_LOADBALANCERV2,
                get_down_bindings=(self.load_balancer.
                                   get_down_loadbalancer_bindings),
                agent_id_attr='agent_id',
                resource_id_attr='loadbalancer_id',
                resource_name='loadbalancer',
                reschedule_resource=self.load_balancer.reschedule_loadbalancer,
                rescheduling_failed=exceptions.LoadbalancerReschedulingFailed)

    def test_loadbalancer_reschedule_from_dead_lbaas_agent(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            loadbalancer_data = loadbalancer['loadbalancer']
            self.plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer_data['id'])
            original_agent = self._get_lbaas_agent_hosting_loadbalancer(
                loadbalancer_data['id'])
            self.assertIsNotNone(original_agent)
            helpers.kill_agent(original_agent['agent']['id'])
            self.load_balancer.reschedule_lbaas_from_down_agents()
            rescheduled_agent = self._get_lbaas_agent_hosting_loadbalancer(
                loadbalancer_data['id'])
            self.assertNotEqual(original_agent, rescheduled_agent)

    def test_reschedule_loadbalancer_succeeded(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            loadbalancer_data = loadbalancer['loadbalancer']
            self.plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer_data['id'])
            hosting_agent = self.load_balancer.get_agent_hosting_loadbalancer(
                self.adminContext, loadbalancer_data['id'])
            with mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager.get_agent_hosting_loadbalancer',
                side_effect=(hosting_agent, hosting_agent)
            ) as mock_get_agent_hosting_lb, mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager._unschedule_loadbalancer',
                side_effect=self.load_balancer._unschedule_loadbalancer
            ) as mock_unschedule_lb, mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager._schedule_loadbalancer',
                side_effect=self.load_balancer._schedule_loadbalancer
            ) as mock_schedule_lb:
                # rescheduling is expected to succeeded
                self.load_balancer.reschedule_loadbalancer(
                    self.adminContext, loadbalancer_data['id'])
                # check the usage of get_agent_hosting_loadbalancer()
                self.assertTrue(mock_get_agent_hosting_lb.called)
                mock_get_agent_hosting_lb.assert_called_with(
                    self.adminContext, loadbalancer_data['id'])
                # check the usage of _unschedule_loadbalancer()
                self.assertTrue(mock_unschedule_lb.called)
                mock_unschedule_lb.assert_called_once_with(
                    self.adminContext, loadbalancer_data['id'],
                    hosting_agent['agent']['id'])
                # check the usage of _schedule_loadbalancer()
                self.assertTrue(mock_schedule_lb.called)
                mock_schedule_lb.assert_called_once_with(
                    self.adminContext, loadbalancer_data['id'])

    def test_reschedule_loadbalancer_failed(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            loadbalancer_data = loadbalancer['loadbalancer']
            self.plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer_data['id'])
            hosting_agent = self.load_balancer.get_agent_hosting_loadbalancer(
                self.adminContext, loadbalancer_data['id'])
            with mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager.get_agent_hosting_loadbalancer',
                side_effect=(hosting_agent, None)
            ) as mock_get_agent_hosting_lb, mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager._unschedule_loadbalancer',
                side_effect=self.load_balancer._unschedule_loadbalancer
            ) as mock_unschedule_lb, mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager._schedule_loadbalancer',
                side_effect=self.load_balancer._schedule_loadbalancer
            ) as mock_schedule_lb:
                # rescheduling is expected to fail
                self.assertRaises(exceptions.LoadbalancerReschedulingFailed,
                                  self.load_balancer.reschedule_loadbalancer,
                                  self.adminContext, loadbalancer_data['id'])
                # check the usage of get_agent_hosting_loadbalancer()
                self.assertTrue(mock_get_agent_hosting_lb.called)
                mock_get_agent_hosting_lb.assert_called_with(
                    self.adminContext, loadbalancer_data['id'])
                # check the usage of _unschedule_loadbalancer()
                self.assertTrue(mock_unschedule_lb.called)
                mock_unschedule_lb.assert_called_once_with(
                    self.adminContext, loadbalancer_data['id'],
                    hosting_agent['agent']['id'])
                # check the usage of _schedule_loadbalancer()
                self.assertTrue(mock_schedule_lb.called)
                mock_schedule_lb.assert_called_once_with(
                    self.adminContext, loadbalancer_data['id'])

    def test__schedule_loadbalancer(self):
        self._register_agent_states(lbaas_agents=True)
        with self.loadbalancer() as loadbalancer:
            loadbalancer_data = loadbalancer['loadbalancer']
            self.plugin.db.update_loadbalancer_provisioning_status(
                self.adminContext, loadbalancer_data['id'])
            with mock.patch(
                'neutron_lbaas.db.loadbalancer.loadbalancer_dbv2.'
                'LoadBalancerPluginDbv2.get_loadbalancer') as mock_get_lb,\
                mock.patch(
                'neutron_lbaas.drivers.common.agent_driver_base.'
                'LoadBalancerManager.create') as mock_create:
                self.load_balancer._schedule_loadbalancer(
                    self.adminContext, loadbalancer_data['id'])
                self.assertTrue(mock_get_lb.called)
                mock_get_lb.assert_called_once_with(self.adminContext,
                                                    loadbalancer_data['id'])
                self.assertTrue(mock_create.called)

    def test__schedule_loadbalancer_already_hosted(self):
        with mock.patch('neutron_lbaas.agent_scheduler.'
                        'LbaasAgentSchedulerDbMixin.'
                        'get_agent_hosting_loadbalancer') as mock_get_agent:
            mock_get_agent.return_value = {'agent': {'id': 'agent-1'}}
            lb = LoadBalancer(id='lb-test')
            agent = self.loadbalancer_scheduler.schedule(
                self.plugin, self.adminContext, lb, self.device_driver)
            self.assertTrue(mock_get_agent.called)
            self.assertIsNone(agent)
